#!/usr/bin/python
# -*- coding: utf-8 -*-

import click
import time
import json
import sys
import subprocess
import os
import webbrowser
from itertools import islice

from .dsl_grammar import *
from .html import html_template_interactive



def listify_and_unify(*args):
    "util to handle listing together dict.keys() sequences"
    out = []
    for x in args:
        if type(x) == list:
            out += x
        else:
            out += list(x)
    return sorted(list(set(out)))


def split_multi_words(llist):
    "break down a list of strings so that it is composed of only 1-word elements"
    broken = [x.split() for x in llist] 
    return list_flatten(broken)

def list_flatten(llist):
    return [item for sublist in llist for item in sublist]

def is_single_word_quoted(w):
    if w[0] == '"' and w[-1] == '"':
        return True
    if w[0] == "'" and w[-1] == "'":
        return True
    return False


def line_last_word(line):
    "return last word"
    if len(line.split()) > 0:
        return line.split()[-1]
    else:
        return ""


def line_last_two_words(line):
    "return last two words"
    if len(line.split()) > 1:
        return " ".join([line.split()[-2], line.split()[-1]])
    else:
        return ""

def line_last_three_words(line):
    "return last three words"
    if len(line.split()) > 2:
        return " ".join([line.split()[-3], line.split()[-2], line.split()[-1]])
    else:
        return ""

def in_square_brackets(line):
    "if we are within the square brackets eg we have a valid 'return <source>[this+that - that' statement"
    if line:
        if line[-1] == "[":
            return True
        else:
            found_opening = False
            for el in reversed(line.split()):
                if "]" in el: 
                    return False
                if "[" in el:
                    found_opening = True
                if el == "return" and found_opening:
                    return True
    return False

def in_categories_search(line):
    """if we are within the `category.name="` bit... """
    pattern = ".name=\""
    if line:
        if line.endswith(pattern):
            for x in G.categories():
                # print("\ntesting", x, "**"+test)
                if line.endswith(x+pattern):
                    # print("found", x)
                    return x
    return False


def line_has_limit_or_skip(line):
    "check if there is a limit or skip statement"
    if "return" in line:
        pos = line.rindex("return") # last position right side
        if "limit" in line[pos:] or "skip" in line[pos:]:
            return True

def line_last_return_subject(line):
    "get the last 'return <object>' statement "
    lista = list(reversed(line.split()))
    # print(lista)
    if len(lista) > 1:
        for el in lista:
            if "return" in el:
                s = lista[lista.index('return') - 1]
                # print(s)
                return s.split("[")[0]


def line_search_subject_is_valid(line):
    "if we have a valid 'search <source>' statement"
    if len(line.split()) == 2 and line.split()[1] in G.sources():
        return True
    else:
        return False

def line_return_subject_is_valid(line):
    "if we have a valid 'return <object>' statement PS not checking if object is semantically valid"
    l = line.split()
    if len(l) > 1:
        if l[-2] == "return":
            return True 

def line_filter_is_partial(line):
    "return one of the valid filter operators after a partial `where -filter-` statement"
    l = line.split()
    if len(l) > 1:
        if l[-2] == "where":
            return True     


def line_filter_is_complete(line):
    "if the filter statement (after where) is fully specified eg `where FOR = '123'` or `where doi!='123'` "
    # @TODO REVIEW
    l = line.split('where')
    if len(l) > 1 and l[-1].strip():
        if 'is empty' in l[-1] or 'is not empty' in l[-1]:
            return True
        else:
            for x in G.lang_filter_operators():
                if x in l[-1]:
                    after_filter = l[-1].split(x) # => [' doi ', ' 123  '] from ' doi = 123  '
                    if len(after_filter) > 1 and after_filter[-1].strip():
                        return True


def line_for_text_search_inner(line):
    "return one of the valid text search operators if inside a `for '...'` statement"
    l = line.split("for")
    if len(l) > 1 and l[-1].strip():
        if l[-1].count("\"") == 1:
            return True

def line_for_text_is_complete(line):
    "if the 'for' text search statement is complete"
    l = line.split("for")
    if len(l) > 1 and l[-1].strip():
        if l[-1].count("\"") > 1 and l[-1].strip()[-1] == "\"":
            return True

def line_is_search_query(line):
    "checks if it is a `search` query"
    l = line.strip().split()
    if len(l) > 1 and l[0] ==  "search":
        return True
    else:
        return False 


def line_search_subject(line):
    "get the source one searches for"
    l = line.split()
    if len(l) > 1 and "search" in l:
        i = l.index("search")
        return l[i + 1]
    else:
        return ""

def line_search_return(line):
    """
    get the source/facet in the return statement
    """
    l = line.split()
    n = l.count("return")
    if n == 1:
        i = l.index("return")
        if len(l) > i + 1: # cause index is zero based
            return_obj = l[i + 1]
            if "[" in return_obj:
                return return_obj.split('[')[0]
            else:
                return return_obj
    else: # if multiple return values, fail
        return None

def line_search_aggregates(line):
    """get the aggregrates statement eg in 
        `search publications return funders aggregate altmetric_median sort by altmetric_median`
        @TODO handle multiple aggregrate statements eg
        `return research_orgs aggregate altmetric_median, rcr_avg sort by rcr_avg`
    """
    l = line.split()
    n = l.count("aggregate")
    if n == 1:
        i = l.index("aggregate")
        if len(l) > i + 2: # cause index is zero based
            return l[i + 1]
    else:
        return None

def line_add_lazy_return(text):
    "if return statement not included, add it lazily"
    if "return" not in text:
        source = line_search_subject(text)
        if source in G.sources():
            # click.secho("..inferring result statement", dim=True)
            return text.strip() + " return " + source
    return text

def line_add_lazy_describe(line):
    "if describe has no arguments, default silently to <describe version>"
    l = line.split()
    if "describe" in line and len(l) == 1:
        return "describe version"
    return line


def save2File(contents, filename, path):
    """Save string contents to a file.

    Not generalized much, so use at your own risk.

    Args:
        contents: string
        filename: string
        path: full valid path

    Returns:
        File location in URL format eg "file://..."
    """
    if not os.path.exists(path):
        os.makedirs(path)
    filename = os.path.join(path, filename)
    f = open(filename, 'wb')
    f.write(contents.encode())  # python will convert \n to os.linesep
    f.close()  # you can omit in most cases as the destructor will call it
    url = "file://" + filename
    return url


def print_warning_prompt_version():
    try:
        from prompt_toolkit import __version__ as prompt_toolkit_version
    except:
        prompt_toolkit_version = "unknown"
    click.secho("WARNING: Dimcli console requires prompt-toolkit version >=2. You are running version '%s'.\nYou can still use Dimcli as Python library. Or upgrade the dependencies with `pip install ipython prompt-toolkit -U" % prompt_toolkit_version,  fg="red")




def open_multi_platform(fpath):
    """
    util to open a file on any platform (i hope)
    """
    click.secho("Opening `%s` ..." % fpath)
    if sys.platform == 'win32':
        subprocess.Popen(['start', fpath], shell=True)

    elif sys.platform == 'darwin':
        subprocess.Popen(['open', fpath])

    else:
        try:
            subprocess.Popen(['xdg-open', fpath])
        except OSError:
            print("Couldnt find suitable opener for %s" % fpath)


def preview_contents(fpath):
    click.secho("File: {}".format(fpath), bold=True)
    try:
        with open(fpath) as f:
            print(f.read())
    except:
        print("An unknown error occured..")


def init_config_folder(user_dir, user_config_file):
    """
    Create the config folder/file unless existing. If it exists, backup and create new one.
    # TODO: update so that it can take KEY based authentication too. 
    # For now it has to be edited manually. 
    """
    if not os.path.exists(user_dir):
        os.mkdir(user_dir)
        # click.secho("Created %s" % user_dir, dim=True)
    if os.path.exists(user_config_file):
        click.secho("Looks like you have already setup a config file: `%s`." % user_config_file, fg="red")
        if click.confirm("Overwrite?"):
            pass
        else:
            click.secho("Goodbye")
            return False

    instance = "[instance.live]" # default for main instance
    url = click.prompt('Please enter the API URL or leave blank for default', default="https://app.dimensions.ai")
    login, password, key = "", "", ""
    if click.confirm('Do you have an API key?'):
        key = click.prompt('Please enter your key', hide_input=True, confirmation_prompt=True)
    else:
        login = click.prompt('Please enter your username')
        password = click.prompt('Please enter your password', hide_input=True, confirmation_prompt=True)
        

    f= open(user_config_file,"w+")
    f.write(instance + "\n")
    f.write("url=" + url + "\n")
    f.write("login=" + login + "\n")
    f.write("password=" + password + "\n")
    f.write("key=" + key + "\n")
    f.close()
    click.secho(
        "Created %s" % user_config_file, dim=True
    )



def init_exports_folder(export_dir):
    """
    Create the folder where json and csv exports are stored
    """
    if not os.path.exists(export_dir):
        os.mkdir(export_dir)


def chunks_of(data, size):
    """Splits up a list or sequence in to chunks of selected size. 

    Args:
        data: A sequence eg a list
        size: A number

    Returns:
        An iterable
    """
    it = iter(data)
    chunk = list(islice(it, size))
    while chunk:
        yield chunk
        chunk = list(islice(it, size))



def exists_key_in_dicts_list(dict_list, key):
    """From a list of dicts, checks if a certain key is in one of the dicts in the list.

    See also https://stackoverflow.com/questions/14790980/how-can-i-check-if-key-exists-in-list-of-dicts-in-python

    Args:
        dict_list: A list of dictionaries
        key: a string to be found in dict keys

    Returns:
        Dictionary or None
    """
    # return next((i for i,d in enumerate(dict_list) if key in d), None)
    return next((d for i,d in enumerate(dict_list) if key in d), None)



def normalize_key(key_name, dict_list, new_val=None):
    """Ensures the key always appear in a JSON dict/objects list, by adding it when missing 
    
    EG
    ```
    for x in pubs_details.publications:
        if not 'FOR' in x:
            x['FOR'] = []
    ```
    becomes
    
    ```
    normalize_key("FOR", pubs_details.publications)
    ```
    Changes happen in-place.

    TIP If `new_val` is not passed, it is inferred from first available non-empty value

    TODO add third argument to pass a lambda function for modifying key
    UPDATE 2019-11-28
    v0.6.1.2: normalizes also 'None' values (to address 1.21 DSL change)
    """
    if new_val == None:
        for x in dict_list:
            if key_name in x:
                new_val = type(x[key_name])() # create empty object eg `list()`
                # print(new_val)
                break 
    for x in dict_list:
        if (not key_name in x) or (x[key_name] == None):
            x[key_name] = new_val



def export_json_csv(jjson, query, USER_EXPORTS_DIR):
    """
    requires the pandas library which is not installed by default

    """
    try:
        from pandas import json_normalize
    except:
        click.secho("This feature requires the pandas library (`pip install pandas` from the terminal)", fg="red")
        return
    return_object = line_search_return(query)
    try:
        df =  json_normalize(jjson[return_object])
    except:
        df =  json_normalize(jjson)
    filename = time.strftime("dsl_export_%Y%m%d-%H%M%S.csv")
    url = save2File(df.to_csv(), filename, USER_EXPORTS_DIR)
    webbrowser.open(url)
    # df.to_csv(USER_EXPORTS_DIR + filename)
    print("Exported: ", "%s%s" % (USER_EXPORTS_DIR, filename))



def export_json_html(jjson, query, api_endpoint, USER_EXPORTS_DIR):
    "print out full json either as pretty_json or within an html template"
    formatted_json = json.dumps(jjson, indent=4, sort_keys=True)
    contents = html_template_interactive(query, formatted_json, api_endpoint)
    filename = time.strftime("%Y%m%d-%H%M%S.html")
    url = save2File(contents, filename, USER_EXPORTS_DIR)
    webbrowser.open(url)
    print("Exported: ", "%s%s" % (USER_EXPORTS_DIR, filename))


def export_json_json(jjson, query, USER_EXPORTS_DIR):
    """
    requires the pandas library which is not installed by default

    """
    formatted_json = json.dumps(jjson, indent=4, sort_keys=True)
    filename = time.strftime("dsl_export_%Y%m%d-%H%M%S.json")
    url = save2File(formatted_json, filename, USER_EXPORTS_DIR)
    webbrowser.open(url)
    # df.to_csv(USER_EXPORTS_DIR + filename)
    print("Exported: ", "%s%s" % (USER_EXPORTS_DIR, filename))



def export_as_bar_chart(jjson, query, USER_EXPORTS_DIR):
    """
    requires the pandas and plotly libraries, which are not installed by default

    """

    try:
        from pandas import json_normalize
        import plotly.express as px
        from plotly.offline import plot
    except:
        click.secho("This feature requires the pandas and plotly library (`pip install pandas plotly` from the terminal)", fg="red")
        return

    return_object = line_search_return(query)
    
    try:
        df =  json_normalize(jjson[return_object])
    except:
        df =  json_normalize(jjson)

    REQUIRED = ["id", "count"]
    for x in REQUIRED:
        if x not in df.columns:
            raise Exception(f"This method requires records to have the keys: {REQUIRED}. It is normally used with facets. Key [{x}] not found.")

    if "title" in df.columns:
        df["value"] = df["id"].astype(str) + df["title"]
    elif "name" in df.columns:
        df["value"] = df["id"].astype(str) + df["name"]
    elif "first_name" in df.columns and "last_name" in df.columns:
        df["value"] = df["id"].astype(str) + df["first_name"] + df["last_name"]
    else: 
        df["value"] = df["id"].astype(str)

    if "country_name" in df.columns:
        color_field = "country_name"
    # if "current_research_org" in df.columns:
    #     color_field = "current_research_org" 
    else:
        color_field = "count"
    
    newplot = px.bar(df, x="value", y="count", title=query, color=color_field)

    filename = time.strftime("dsl_plot_export_%Y%m%d-%H%M%S.html")
    plot(newplot, filename = USER_EXPORTS_DIR+filename, auto_open=True)

    print("Exported: ", "%s%s" % (USER_EXPORTS_DIR, filename))




def export_as_jupyter(jjson, query, USER_EXPORTS_DIR):
    """
    Take the last 10 (default) rows from the history, and create a new python notebook with them. 
    Saves in usual location.

    Based on https://gist.github.com/fperez/9716279
    """

    try:
        from itertools import islice
        import subprocess
        import nbformat as nbf
    except:
        click.secho("This feature requires the nbformat library (`pip install nbformat` from the terminal)", fg="red")
        return

    nb = nbf.v4.new_notebook()
    this_time = time.strftime("%Y.%m.%d h%H:%M:%S")

    text = f"""# Dimensions API Queries Export\n### {this_time} \nThis notebook was generated using [Dimcli](https://github.com/digital-science/dimcli/) - the Dimensions API CLI."""

    nb['cells'] = [nbf.v4.new_markdown_cell(text)]

    setup = """!pip install dimcli -U --quiet\nimport dimcli\ndimcli.login()"""
    nb['cells'] += [nbf.v4.new_code_cell(setup)]

    nb['cells'] += [nbf.v4.new_code_cell("%%dsl\n" + query)]

    formatted_json = json.dumps(jjson, indent=4, sort_keys=True)
    results_preview = f"""### Precomputed Query Results: \n```\n{formatted_json}\n```"""
    nb['cells'] += [nbf.v4.new_markdown_cell(results_preview)]

    filename = time.strftime(f"dsl_export_%Y-%m-%d_%H-%M-%S.ipynb")
    nbf.write(nb, USER_EXPORTS_DIR + filename)
    
    subprocess.run(['open', USER_EXPORTS_DIR + filename], check=True)
    print("Exported: ", "%s%s" % (USER_EXPORTS_DIR, filename))





def print_json_stats(res, query=""):
    """
    from a dimcli.DslDataset object, print out basic stats
    * works primarily for 'search' types of query
    """
    # what is searched for
    source, tot = line_search_subject(query), None
    if source:
        if res['stats']:
            tot = res['stats']["total_count"]
        for k in res.good_data_keys():
            if tot and source == k:
                print(f"Returned {source.capitalize()}: {len(res[source])} (total = {tot})")
            else:
                print(f"Returned {k.capitalize()}: {len(res[k])}")



def print_json_errors(res):
    """
    from a dimcli.DslDataset object, print out an errors summary
    """
    if "errors" in res.json.keys():
        if "query" in res.json["errors"]:
            print(res.json["errors"]["query"]["header"].strip("\n"))
            for key in res.json["errors"]["query"]["details"]:
                print(key)
        else:
            print(res.json["errors"])

def print_json_warnings(res):
    """
    from a dimcli.DslDataset object, print out a warnings
    """
    if "_warnings" in res.json.keys():
        print("WARNINGS [{}]".format(len(res.json["_warnings"])))
        print("\n".join([s for s in res.json["_warnings"]]))



def preview_results(jsondata, maxitems=10):
    """
    Preview items in console
    If it's one of the main sources, try to show title/id. Otherwise show json in one line
    """
    # click.secho("Showing first %d records from latest query.." % maxitems, dim=True)
    # click.secho("")
    counter = 0
    for key in jsondata.keys():
        if key not in ["_stats", "_warnings", "_notes",  "_version"]:
            for row in jsondata[key]:
                counter += 1
                if counter <= maxitems:
                    try:  # title and url/id if object has them
                        url = dimensions_url(row['id'], key, verbose=False) or row['id']
                        if 'title' in row.keys(): # for docs
                            name_or_title = row['title'].strip()
                        elif 'name' in row.keys():  # for orgs
                            name_or_title = row['name'].strip()
                        else: # for researchers
                            name_or_title = row['first_name'] + " " + row['last_name']
                        click.echo(
                            click.style("[" + str(counter) + "] ", dim=True) +
                            click.style(name_or_title , bold=True) +
                            click.style(" (id: " + url + " )", fg='blue'))

                    except:  # fallback: full row
                        click.echo(
                            click.style("[" + str(counter) + "] ", dim=True) +
                            click.style(str(row)))
            if False:
                click.secho("---", dim=True)
                click.secho("Tip: use 'show <number>' or show+Tab to see more options", dim=True)



def print_json_compact(jsondata):
    """
    Show json in one line
    NOTE: the logic is the same as the except clause in print_json_preview. Maybe some refactoring could be beneficial here..
    """
    from pygments import highlight, lexers, formatters
    lexer = lexers.JsonLexer()
    formatter = formatters.TerminalFormatter()
    
    counter = 0
    for key in jsondata.keys():
        if key == "_stats":
            pass
        else:
            for row in jsondata[key]:
                counter += 1
                # full row
                click.echo(
                    click.style("[" + str(counter) + "] ", dim=True) +
                    click.style(highlight(json.dumps(row, sort_keys=True).strip(), lexer, formatter).strip()))



def print_json_full(jsondata):
    """
    pretty print json 
    """
    formatted_json = json.dumps(jsondata, indent=4, sort_keys=True)
    from pygments import highlight, lexers, formatters
    colorful_json = highlight(formatted_json, lexers.JsonLexer(),
                                formatters.TerminalFormatter())
    print(colorful_json)




def dimensions_url(obj_id, obj_type="", verbose=True):
    """
    Generate a valid Dimensions URL for one of the available sources.
    obj_id: the dimensions ID of the object
    obj_type: one of 'publications', 'grants', 'patents', 'policy_documents', 'clinical_trials', 'researchers'
    """
    
    if obj_type and obj_type not in G.sources():
        raise ValueError("ERROR: valid sources are: " + " ".join([x for x in G.sources()]))
    if not obj_type:
        for source, prefix in G.object_id_patterns().items():
            if obj_id.startswith(prefix):
                obj_type = source
    if obj_type:
        url = G.url_for_source(obj_type)
        if url:
            return url + obj_id


def print_dimensions_url(obj_id):
    "wrapper dimensions_url() - for use within the cli"
    url = dimensions_url(obj_id)
    if url:
        print("Got a match:")
        print(url)
    else:
        # in this case, it's a cltrial or patent
        print("Cannot resolve automatically. Maybe:")
        print(dimensions_url(obj_id, "patents"))
        print(dimensions_url(obj_id, "clinical_trials"))



def dimensions_url_search(keywords_list_as_string):
    "Returns a valid keyword search URL for Dimensions"
    q = """https://app.dimensions.ai/discover/publication?search_text={}&search_type=kws&search_field=full_search"""
    from urllib.parse import quote   
    s = quote(keywords_list_as_string)  
    return q.format(s)


def google_url(stringa):
    """
    Generate a valid google search URL from a string (URL quoting is applied)
    """
    from urllib.parse import quote   
    s = quote(stringa)    
    return f"https://www.google.com/search?q={s}"
 


def dsl_escape(stringa, all=False):   
    """
    Helper for escaping the full-text inner query string, when it includes quotes. Usage:

    `search publications for "{dsl_escape(complex_q)}" return publications`

    EG imagine the query string:
    '"2019-nCoV" OR "COVID-19" OR "SARS-CoV-2" OR (("coronavirus"  OR "corona virus") AND (Wuhan OR China))'
    
    In Python, if you want to embed it into a DSL query, it has to become:
    '\\"2019-nCoV\\" OR \\"COVID-19\\" OR \\"SARS-CoV-2\\" OR ((\\"coronavirus\\"  OR \\"corona virus\\") AND (Wuhan OR China))'

    NOTE by default only quotes as escaped. If you want to escape all special chars, pass all=True, eg

    > dsl_escape('Solar cells: a new technology?', True)
    > 'Solar cells\\: a new technology?'

    See also: https://docs.dimensions.ai/dsl/language.html#for-search-term
    """
    
    if all:
        escaped = stringa.translate(str.maketrans({"^":  r"\^",
                                                    '"':  r'\"',
                                                    "\\": r"\\",
                                                    ":":  r"\:",
                                                    "~":  r"\~",
                                                    "[":  r"\[",
                                                    "]":  r"\]",
                                                    "{":  r"\{",
                                                    "}":  r"\}",
                                                    "(":  r"\(",
                                                    ")":  r"\)",
                                                    "!":  r"\!",
                                                    "|":  r"\|",
                                                    "&":  r"\&",
                                                    "+":  r"\+",
                                                    }))
    else:
        escaped = stringa.translate(str.maketrans({'"':  r'\"'}))        
    return escaped



#!/usr/bin/python
# -*- coding: utf-8 -*-


#
# CATEGORIES_DICT is a dictionary rendering of the DSL research categories JSON
#
# last updated: 2019-09-23
#
# how to create:
#
# > search publications return category_for limit 1000
# (for each of the categories)
#
# Then remove '_stats' section, and combine them all into a py file
#
#

CATEGORIES_DICT = {
    "category_sdg": [
        {
            "count": 1348219,
            "id": "40007",
            "name": "7 Affordable and Clean Energy"
        },
        {
            "count": 1265880,
            "id": "40003",
            "name": "3 Good Health and Well Being"
        },
        {
            "count": 641929,
            "id": "40016",
            "name": "16 Peace, Justice and Strong Institutions"
        },
        {
            "count": 515160,
            "id": "40004",
            "name": "4 Quality Education"
        },
        {
            "count": 429531,
            "id": "40013",
            "name": "13 Climate Action"
        },
        {
            "count": 227782,
            "id": "40008",
            "name": "8 Decent Work and Economic Growth"
        },
        {
            "count": 195027,
            "id": "40011",
            "name": "11 Sustainable Cities and Communities"
        },
        {
            "count": 164207,
            "id": "40010",
            "name": "10 Reduced Inequalities"
        },
        {
            "count": 119014,
            "id": "40002",
            "name": "2 Zero Hunger"
        },
        {
            "count": 70387,
            "id": "40006",
            "name": "6 Clean Water and Sanitation"
        },
        {
            "count": 61860,
            "id": "40014",
            "name": "14 Life Below Water"
        },
        {
            "count": 61217,
            "id": "40015",
            "name": "15 Life on Land"
        },
        {
            "count": 57750,
            "id": "40012",
            "name": "12 Responsible Consumption and Production"
        },
        {
            "count": 43545,
            "id": "40001",
            "name": "1 No Poverty"
        },
        {
            "count": 21476,
            "id": "40005",
            "name": "5 Gender Equality"
        },
        {
            "count": 19311,
            "id": "40009",
            "name": "9 Industry, Innovation and Infrastructure"
        },
        {
            "count": 10346,
            "id": "40017",
            "name": "17 Partnerships for the Goals"
        }
    ],
    "category_uoa": [
        {
            "count": 12965896,
            "id": "30012",
            "name": "B12 Engineering"
        },
        {
            "count": 7664934,
            "id": "30001",
            "name": "A01 Clinical Medicine"
        },
        {
            "count": 5441205,
            "id": "30003",
            "name": "A03 Allied Health Professions, Dentistry, Nursing and Pharmacy"
        },
        {
            "count": 3476541,
            "id": "30011",
            "name": "B11 Computer Science and Informatics"
        },
        {
            "count": 2674165,
            "id": "30008",
            "name": "B08 Chemistry"
        },
        {
            "count": 2589803,
            "id": "30005",
            "name": "A05 Biological Sciences"
        },
        {
            "count": 2504503,
            "id": "30009",
            "name": "B09 Physics"
        },
        {
            "count": 2361360,
            "id": "30004",
            "name": "A04 Psychology, Psychiatry and Neuroscience"
        },
        {
            "count": 2348396,
            "id": "30010",
            "name": "B10 Mathematical Sciences"
        },
        {
            "count": 2260339,
            "id": "30006",
            "name": "A06 Agriculture, Veterinary and Food Science"
        },
        {
            "count": 2023701,
            "id": "30017",
            "name": "C17 Business and Management Studies"
        },
        {
            "count": 1618104,
            "id": "30007",
            "name": "B07 Earth Systems and Environmental Sciences"
        },
        {
            "count": 960452,
            "id": "30023",
            "name": "C23 Education"
        },
        {
            "count": 840151,
            "id": "30002",
            "name": "A02 Public Health, Health Services and Primary Care"
        },
        {
            "count": 662202,
            "id": "30014",
            "name": "C14 Geography and Environmental Studies"
        },
        {
            "count": 467677,
            "id": "30028",
            "name": "D28 History"
        },
        {
            "count": 447784,
            "id": "30024",
            "name": "C24 Sport and Exercise Sciences, Leisure and Tourism"
        },
        {
            "count": 398831,
            "id": "30013",
            "name": "C13 Architecture, Built Environment and Planning"
        },
        {
            "count": 389544,
            "id": "30020",
            "name": "C20 Social Work and Social Policy"
        },
        {
            "count": 380338,
            "id": "30018",
            "name": "C18 Law"
        },
        {
            "count": 355463,
            "id": "30019",
            "name": "C19 Politics and International Studies"
        },
        {
            "count": 311640,
            "id": "30026",
            "name": "D26 Modern Languages and Linguistics"
        },
        {
            "count": 275604,
            "id": "30016",
            "name": "C16 Economics and Econometrics"
        },
        {
            "count": 251384,
            "id": "30027",
            "name": "D27 English Language and Literature"
        },
        {
            "count": 200627,
            "id": "30030",
            "name": "D30 Philosophy"
        },
        {
            "count": 195652,
            "id": "30034",
            "name": "D34 Communication, Cultural and Media Studies, Library and Information Management"
        },
        {
            "count": 133441,
            "id": "30032",
            "name": "D32 Art and Design: History, Practice and Theory"
        },
        {
            "count": 130495,
            "id": "30022",
            "name": "C22 Anthropology and Development Studies"
        },
        {
            "count": 125591,
            "id": "30021",
            "name": "C21 Sociology"
        },
        {
            "count": 92204,
            "id": "30033",
            "name": "D33 Music, Drama, Dance, Performing Arts, Film and Screen Studies"
        },
        {
            "count": 87084,
            "id": "30015",
            "name": "C15 Archaeology"
        },
        {
            "count": 76316,
            "id": "30031",
            "name": "D31 Theology and Religious Studies"
        },
        {
            "count": 71109,
            "id": "30025",
            "name": "D25 Area Studies"
        },
        {
            "count": 24820,
            "id": "30029",
            "name": "D29 Classics"
        }
    ],
    "category_rcdc": [
        {
            "count": 6974252,
            "id": "507",
            "name": "Clinical Research"
        },
        {
            "count": 2977663,
            "id": "558",
            "name": "Prevention"
        },
        {
            "count": 2960147,
            "id": "387",
            "name": "Neurosciences"
        },
        {
            "count": 2879104,
            "id": "503",
            "name": "Cancer"
        },
        {
            "count": 2717263,
            "id": "526",
            "name": "Genetics"
        },
        {
            "count": 2621641,
            "id": "559",
            "name": "Rare Diseases"
        },
        {
            "count": 2253835,
            "id": "547",
            "name": "Pediatric"
        },
        {
            "count": 1984380,
            "id": "501",
            "name": "Brain Disorders"
        },
        {
            "count": 1943567,
            "id": "533",
            "name": "Infectious Diseases"
        },
        {
            "count": 1917560,
            "id": "498",
            "name": "Behavioral and Social Science"
        },
        {
            "count": 1845002,
            "id": "344",
            "name": "Digestive Diseases"
        },
        {
            "count": 1783610,
            "id": "331",
            "name": "Cardiovascular"
        },
        {
            "count": 1559869,
            "id": "337",
            "name": "Bioengineering"
        },
        {
            "count": 1494264,
            "id": "338",
            "name": "Biotechnology"
        },
        {
            "count": 1193630,
            "id": "388",
            "name": "Nutrition"
        },
        {
            "count": 1171906,
            "id": "295",
            "name": "Aging"
        },
        {
            "count": 1153989,
            "id": "380",
            "name": "Mental Health"
        },
        {
            "count": 1078591,
            "id": "445",
            "name": "Heart Disease"
        },
        {
            "count": 1009901,
            "id": "508",
            "name": "Clinical Trials and Supportive Activities"
        },
        {
            "count": 948782,
            "id": "410",
            "name": "Lung"
        },
        {
            "count": 859626,
            "id": "447",
            "name": "Hematology"
        },
        {
            "count": 854776,
            "id": "542",
            "name": "Neurodegenerative"
        },
        {
            "count": 804596,
            "id": "520",
            "name": "Emerging Infectious Diseases"
        },
        {
            "count": 783143,
            "id": "439",
            "name": "Diagnostic Radiology"
        },
        {
            "count": 753280,
            "id": "546",
            "name": "Patient Safety"
        },
        {
            "count": 746542,
            "id": "583",
            "name": "Vaccine Related"
        },
        {
            "count": 660911,
            "id": "451",
            "name": "Liver Disease"
        },
        {
            "count": 659628,
            "id": "497",
            "name": "Basic Behavioral and Social Science"
        },
        {
            "count": 567077,
            "id": "397",
            "name": "Pain Research"
        },
        {
            "count": 562711,
            "id": "443",
            "name": "Health Services"
        },
        {
            "count": 534635,
            "id": "537",
            "name": "Kidney Disease"
        },
        {
            "count": 515608,
            "id": "438",
            "name": "Diabetes"
        },
        {
            "count": 513807,
            "id": "496",
            "name": "Autoimmune Disease"
        },
        {
            "count": 509504,
            "id": "521",
            "name": "Eye Disease And Disorders Of Vision"
        },
        {
            "count": 497532,
            "id": "436",
            "name": "Dental/Oral and Craniofacial Disease"
        },
        {
            "count": 473762,
            "id": "573",
            "name": "Substance Abuse"
        },
        {
            "count": 450652,
            "id": "416",
            "name": "Contraception/Reproduction"
        },
        {
            "count": 437544,
            "id": "484",
            "name": "Stem Cell Research"
        },
        {
            "count": 422939,
            "id": "421",
            "name": "Pain Conditions - Chronic"
        },
        {
            "count": 422080,
            "id": "430",
            "name": "Complementary and Alternative Medicine"
        },
        {
            "count": 407827,
            "id": "314",
            "name": "Injury (total) Accidents/Adverse Effects"
        },
        {
            "count": 406674,
            "id": "363",
            "name": "Human Genome"
        },
        {
            "count": 404513,
            "id": "580",
            "name": "Urologic Diseases"
        },
        {
            "count": 359381,
            "id": "527",
            "name": "HIV/AIDS"
        },
        {
            "count": 356825,
            "id": "499",
            "name": "Biodefense"
        },
        {
            "count": 356172,
            "id": "444",
            "name": "Heart Disease - Coronary Heart Disease"
        },
        {
            "count": 352817,
            "id": "316",
            "name": "Breast Cancer"
        },
        {
            "count": 345878,
            "id": "469",
            "name": "Rehabilitation"
        },
        {
            "count": 344591,
            "id": "578",
            "name": "Transplantation"
        },
        {
            "count": 337166,
            "id": "389",
            "name": "Obesity"
        },
        {
            "count": 328466,
            "id": "465",
            "name": "Perinatal Period - Conditions Originating in Perinatal Period"
        },
        {
            "count": 314573,
            "id": "390",
            "name": "Nanotechnology"
        },
        {
            "count": 312027,
            "id": "308",
            "name": "Atherosclerosis"
        },
        {
            "count": 303111,
            "id": "531",
            "name": "Immunization"
        },
        {
            "count": 281733,
            "id": "437",
            "name": "Depression"
        },
        {
            "count": 279118,
            "id": "375",
            "name": "Chronic Liver Disease and Cirrhosis"
        },
        {
            "count": 266299,
            "id": "379",
            "name": "Regenerative Medicine"
        },
        {
            "count": 260234,
            "id": "543",
            "name": "Orphan Drug"
        },
        {
            "count": 243219,
            "id": "458",
            "name": "Mind and Body"
        },
        {
            "count": 224522,
            "id": "304",
            "name": "Arthritis"
        },
        {
            "count": 216988,
            "id": "487",
            "name": "Acquired Cognitive Impairment"
        },
        {
            "count": 213537,
            "id": "485",
            "name": "Stroke"
        },
        {
            "count": 210929,
            "id": "367",
            "name": "Hypertension"
        },
        {
            "count": 205852,
            "id": "517",
            "name": "Drug Abuse (NIDA Only)"
        },
        {
            "count": 203263,
            "id": "569",
            "name": "Stem Cell Research - Nonembryonic - Non-Human"
        },
        {
            "count": 201281,
            "id": "368",
            "name": "Infant Mortality/ (LBW)"
        },
        {
            "count": 197817,
            "id": "514",
            "name": "Dementia"
        },
        {
            "count": 194044,
            "id": "305",
            "name": "Assistive Technology"
        },
        {
            "count": 185634,
            "id": "427",
            "name": "Vector-Borne Diseases"
        },
        {
            "count": 185247,
            "id": "490",
            "name": "Alcoholism, Alcohol Use and Health"
        },
        {
            "count": 180810,
            "id": "414",
            "name": "Colo-Rectal Cancer"
        },
        {
            "count": 180592,
            "id": "302",
            "name": "Antimicrobial Resistance"
        },
        {
            "count": 176919,
            "id": "541",
            "name": "Networking and Information Technology R&D"
        },
        {
            "count": 175828,
            "id": "550",
            "name": "Pediatric Research Initiative"
        },
        {
            "count": 169271,
            "id": "326",
            "name": "Lung Cancer"
        },
        {
            "count": 161735,
            "id": "556",
            "name": "Pneumonia & Influenza"
        },
        {
            "count": 161079,
            "id": "448",
            "name": "Hepatitis"
        },
        {
            "count": 158223,
            "id": "493",
            "name": "Alzheimer's Disease including Alzheimer's Disease Related Dementias (AD/ADRD)"
        },
        {
            "count": 150163,
            "id": "568",
            "name": "Stem Cell Research - Nonembryonic - Human"
        },
        {
            "count": 149741,
            "id": "298",
            "name": "Alzheimer's Disease"
        },
        {
            "count": 143283,
            "id": "478",
            "name": "Sleep Research"
        },
        {
            "count": 140093,
            "id": "536",
            "name": "Intellectual and Developmental Disabilities (IDD)"
        },
        {
            "count": 139955,
            "id": "412",
            "name": "Tobacco"
        },
        {
            "count": 138102,
            "id": "480",
            "name": "Smoking and Health"
        },
        {
            "count": 132449,
            "id": "463",
            "name": "Perinatal - Birth - Preterm (LBW)"
        },
        {
            "count": 129607,
            "id": "377",
            "name": "Prostate Cancer"
        },
        {
            "count": 129360,
            "id": "353",
            "name": "Estrogen"
        },
        {
            "count": 126029,
            "id": "454",
            "name": "Lymphoma"
        },
        {
            "count": 125504,
            "id": "391",
            "name": "Organ Transplantation"
        },
        {
            "count": 120129,
            "id": "588",
            "name": "Violence Research"
        },
        {
            "count": 120119,
            "id": "522",
            "name": "Foodborne Illness"
        },
        {
            "count": 116438,
            "id": "561",
            "name": "Serious Mental Illness"
        },
        {
            "count": 113981,
            "id": "476",
            "name": "Sexually Transmitted Diseases/Herpes"
        },
        {
            "count": 107221,
            "id": "394",
            "name": "Osteoporosis"
        },
        {
            "count": 106167,
            "id": "422",
            "name": "Tuberculosis"
        },
        {
            "count": 104341,
            "id": "324",
            "name": "Injury - Trauma - (Head and Spine)"
        },
        {
            "count": 102292,
            "id": "352",
            "name": "Epilepsy"
        },
        {
            "count": 99389,
            "id": "426",
            "name": "Pneumonia"
        },
        {
            "count": 97891,
            "id": "306",
            "name": "Asthma"
        },
        {
            "count": 97449,
            "id": "313",
            "name": "Brain Cancer"
        },
        {
            "count": 95996,
            "id": "472",
            "name": "Schizophrenia"
        },
        {
            "count": 95096,
            "id": "549",
            "name": "Pediatric Cancer"
        },
        {
            "count": 91389,
            "id": "509",
            "name": "Comparative Effectiveness Research"
        },
        {
            "count": 83120,
            "id": "396",
            "name": "Ovarian Cancer"
        },
        {
            "count": 82088,
            "id": "411",
            "name": "Pancreatic Cancer"
        },
        {
            "count": 79078,
            "id": "511",
            "name": "Congenital Structural Anomalies"
        },
        {
            "count": 78257,
            "id": "553",
            "name": "Peripheral Neuropathy"
        },
        {
            "count": 73962,
            "id": "525",
            "name": "Genetic Testing"
        },
        {
            "count": 72209,
            "id": "398",
            "name": "Parkinson's Disease"
        },
        {
            "count": 71018,
            "id": "450",
            "name": "Liver Cancer"
        },
        {
            "count": 70947,
            "id": "369",
            "name": "Inflammatory Bowel Disease"
        },
        {
            "count": 68773,
            "id": "475",
            "name": "Septicemia"
        },
        {
            "count": 68644,
            "id": "317",
            "name": "Cervical Cancer"
        },
        {
            "count": 65909,
            "id": "370",
            "name": "Influenza"
        },
        {
            "count": 61919,
            "id": "429",
            "name": "Climate-Related Exposures and Conditions"
        },
        {
            "count": 60835,
            "id": "460",
            "name": "Multiple Sclerosis"
        },
        {
            "count": 60738,
            "id": "409",
            "name": "Infertility"
        },
        {
            "count": 60613,
            "id": "471",
            "name": "Rural Health"
        },
        {
            "count": 59525,
            "id": "457",
            "name": "Malaria"
        },
        {
            "count": 59016,
            "id": "482",
            "name": "Spinal Cord Injury"
        },
        {
            "count": 56647,
            "id": "452",
            "name": "Lupus"
        },
        {
            "count": 56618,
            "id": "419",
            "name": "Hepatitis - C"
        },
        {
            "count": 56124,
            "id": "361",
            "name": "Gene Therapy"
        },
        {
            "count": 52331,
            "id": "506",
            "name": "Childhood Leukemia"
        },
        {
            "count": 51837,
            "id": "554",
            "name": "Physical Rehabilitation"
        },
        {
            "count": 46273,
            "id": "418",
            "name": "Hepatitis - B"
        },
        {
            "count": 45934,
            "id": "310",
            "name": "Autism"
        },
        {
            "count": 42660,
            "id": "376",
            "name": "Chronic Obstructive Pulmonary Disease"
        },
        {
            "count": 41391,
            "id": "519",
            "name": "Emergency Care"
        },
        {
            "count": 40592,
            "id": "494",
            "name": "Anxiety Disorders"
        },
        {
            "count": 40334,
            "id": "433",
            "name": "Cost Effectiveness Research"
        },
        {
            "count": 40331,
            "id": "510",
            "name": "Conditions Affecting the Embryonic and Fetal Periods"
        },
        {
            "count": 39101,
            "id": "400",
            "name": "Suicide"
        },
        {
            "count": 38058,
            "id": "435",
            "name": "Cystic Fibrosis"
        },
        {
            "count": 36415,
            "id": "513",
            "name": "Crohn's Disease"
        },
        {
            "count": 36313,
            "id": "325",
            "name": "Injury - Traumatic brain injury"
        },
        {
            "count": 34164,
            "id": "442",
            "name": "Headaches"
        },
        {
            "count": 30626,
            "id": "557",
            "name": "Post-Traumatic Stress Disorder (PTSD)"
        },
        {
            "count": 30501,
            "id": "345",
            "name": "Digestive Diseases - (Peptic Ulcer)"
        },
        {
            "count": 29193,
            "id": "560",
            "name": "Rheumatoid Arthritis"
        },
        {
            "count": 28216,
            "id": "349",
            "name": "Eating Disorders"
        },
        {
            "count": 26993,
            "id": "589",
            "name": "Youth Violence"
        },
        {
            "count": 26749,
            "id": "309",
            "name": "Attention Deficit Disorder (ADD)"
        },
        {
            "count": 26356,
            "id": "564",
            "name": "Stem Cell Research - Embryonic - Non-Human"
        },
        {
            "count": 25721,
            "id": "575",
            "name": "Suicide Prevention"
        },
        {
            "count": 24320,
            "id": "548",
            "name": "Pediatric AIDS"
        },
        {
            "count": 24278,
            "id": "383",
            "name": "Muscular Dystrophy"
        },
        {
            "count": 24131,
            "id": "455",
            "name": "Macular Degeneration"
        },
        {
            "count": 23796,
            "id": "534",
            "name": "Injury - Childhood Injuries"
        },
        {
            "count": 23233,
            "id": "488",
            "name": "Adolescent Sexual Activity"
        },
        {
            "count": 22804,
            "id": "293",
            "name": "Acute Respiratory Distress Syndrome"
        },
        {
            "count": 22486,
            "id": "378",
            "name": "Psoriasis"
        },
        {
            "count": 22437,
            "id": "322",
            "name": "Migraines"
        },
        {
            "count": 20413,
            "id": "311",
            "name": "Cerebral Palsy"
        },
        {
            "count": 20206,
            "id": "587",
            "name": "Violence Against Women"
        },
        {
            "count": 20113,
            "id": "582",
            "name": "Uterine Cancer"
        },
        {
            "count": 20033,
            "id": "544",
            "name": "Osteoarthritis"
        },
        {
            "count": 17266,
            "id": "477",
            "name": "Sickle Cell Disease"
        },
        {
            "count": 17232,
            "id": "292",
            "name": "ALS"
        },
        {
            "count": 16739,
            "id": "332",
            "name": "Child Abuse and Neglect Research"
        },
        {
            "count": 16005,
            "id": "563",
            "name": "Stem Cell Research - Embryonic - Human"
        },
        {
            "count": 15973,
            "id": "395",
            "name": "Otitis Media"
        },
        {
            "count": 15786,
            "id": "381",
            "name": "Neuroblastoma"
        },
        {
            "count": 15727,
            "id": "359",
            "name": "Food Allergies"
        },
        {
            "count": 15649,
            "id": "403",
            "name": "Temporomandibular Muscle/Joint Disorder (TMJD)"
        },
        {
            "count": 14002,
            "id": "321",
            "name": "Methamphetamine"
        },
        {
            "count": 13842,
            "id": "424",
            "name": "Underage Drinking"
        },
        {
            "count": 13760,
            "id": "351",
            "name": "Endometriosis"
        },
        {
            "count": 13721,
            "id": "565",
            "name": "Stem Cell Research - Induced Pluripotent Stem Cell"
        },
        {
            "count": 13131,
            "id": "473",
            "name": "Scleroderma"
        },
        {
            "count": 12860,
            "id": "562",
            "name": "Sexual and Gender Minorities (SGM/LGBT*)"
        },
        {
            "count": 12655,
            "id": "300",
            "name": "Anorexia"
        },
        {
            "count": 11301,
            "id": "505",
            "name": "Cannabinoid Research"
        },
        {
            "count": 11220,
            "id": "364",
            "name": "Huntington's Disease"
        },
        {
            "count": 11112,
            "id": "577",
            "name": "Transmissible Spongiform Encephalopathy (TSE)"
        },
        {
            "count": 10894,
            "id": "566",
            "name": "Stem Cell Research - Induced Pluripotent Stem Cell - Human"
        },
        {
            "count": 10734,
            "id": "299",
            "name": "American Indians / Alaska Natives"
        },
        {
            "count": 10571,
            "id": "303",
            "name": "Aphasia"
        },
        {
            "count": 10313,
            "id": "365",
            "name": "Hydrocephalus"
        },
        {
            "count": 9713,
            "id": "441",
            "name": "HPV and/or Cervical Cancer Vaccines"
        },
        {
            "count": 9698,
            "id": "461",
            "name": "Myasthenia Gravis"
        },
        {
            "count": 9673,
            "id": "468",
            "name": "Prescription Drug Abuse"
        },
        {
            "count": 9386,
            "id": "347",
            "name": "Duchenne/ Becker Muscular Dystrophy"
        },
        {
            "count": 9362,
            "id": "532",
            "name": "Infant Mortality"
        },
        {
            "count": 9252,
            "id": "584",
            "name": "Vaccine related (AIDS)"
        },
        {
            "count": 9083,
            "id": "346",
            "name": "Down Syndrome"
        },
        {
            "count": 9050,
            "id": "570",
            "name": "Stem Cell Research - Umbilical Cord Blood/ Placenta"
        },
        {
            "count": 8908,
            "id": "453",
            "name": "Lyme Disease"
        },
        {
            "count": 8515,
            "id": "350",
            "name": "Emphysema"
        },
        {
            "count": 8472,
            "id": "539",
            "name": "Major Depressive Disorder"
        },
        {
            "count": 8328,
            "id": "500",
            "name": "Bipolar Disorder"
        },
        {
            "count": 8140,
            "id": "362",
            "name": "Homelessness"
        },
        {
            "count": 8099,
            "id": "552",
            "name": "Perinatal - Neonatal Respiratory Distress Syndrome"
        },
        {
            "count": 7921,
            "id": "535",
            "name": "Injury - Unintentional Childhood Injury"
        },
        {
            "count": 7892,
            "id": "516",
            "name": "Digestive Diseases - (Gallbladder)"
        },
        {
            "count": 7632,
            "id": "512",
            "name": "Cooley's Anemia"
        },
        {
            "count": 7234,
            "id": "467",
            "name": "Polycystic Kidney Disease"
        },
        {
            "count": 7167,
            "id": "358",
            "name": "Fibromyalgia"
        },
        {
            "count": 7081,
            "id": "491",
            "name": "Allergic Rhinitis (Hay Fever)"
        },
        {
            "count": 6849,
            "id": "348",
            "name": "Dystonia"
        },
        {
            "count": 6814,
            "id": "385",
            "name": "Neurofibromatosis"
        },
        {
            "count": 6806,
            "id": "571",
            "name": "Stem Cell Research - Umbilical Cord Blood/ Placenta - Human"
        },
        {
            "count": 6701,
            "id": "523",
            "name": "Frontotemporal Dementia (FTD)"
        },
        {
            "count": 6137,
            "id": "492",
            "name": "Alzheimer's Disease Related Dementias (ADRD)"
        },
        {
            "count": 5882,
            "id": "529",
            "name": "Health Effects of Indoor Air Pollution"
        },
        {
            "count": 5530,
            "id": "402",
            "name": "Teenage Pregnancy"
        },
        {
            "count": 5465,
            "id": "590",
            "name": "Youth Violence Prevention"
        },
        {
            "count": 5451,
            "id": "405",
            "name": "West Nile Virus"
        },
        {
            "count": 4903,
            "id": "489",
            "name": "Agent Orange & Dioxin"
        },
        {
            "count": 4885,
            "id": "301",
            "name": "Anthrax"
        },
        {
            "count": 4596,
            "id": "341",
            "name": "Fragile X Syndrome"
        },
        {
            "count": 4178,
            "id": "357",
            "name": "Fibroid Tumors (Uterine)"
        },
        {
            "count": 3871,
            "id": "423",
            "name": "Tuberous Sclerosis"
        },
        {
            "count": 3825,
            "id": "374",
            "name": "Chronic Fatigue Syndrome (ME/CFS)"
        },
        {
            "count": 3803,
            "id": "328",
            "name": "Interstitial Cystitis"
        },
        {
            "count": 3750,
            "id": "318",
            "name": "Charcot-Marie-Tooth Disease"
        },
        {
            "count": 3689,
            "id": "502",
            "name": "Burden of Illness"
        },
        {
            "count": 3685,
            "id": "579",
            "name": "Tuberculosis Vaccine"
        },
        {
            "count": 3577,
            "id": "481",
            "name": "Spina Bifida"
        },
        {
            "count": 3567,
            "id": "420",
            "name": "Hodgkin's Disease"
        },
        {
            "count": 3418,
            "id": "486",
            "name": "Substance Abuse Prevention"
        },
        {
            "count": 3400,
            "id": "574",
            "name": "Sudden Infant Death Syndrome"
        },
        {
            "count": 3313,
            "id": "356",
            "name": "Fetal Alcohol Syndrome"
        },
        {
            "count": 3307,
            "id": "479",
            "name": "Small Pox"
        },
        {
            "count": 3100,
            "id": "393",
            "name": "Osteogenesis Imperfecta"
        },
        {
            "count": 3080,
            "id": "330",
            "name": "Tourette Syndrome"
        },
        {
            "count": 2891,
            "id": "540",
            "name": "Myotonic Dystrophy"
        },
        {
            "count": 2824,
            "id": "307",
            "name": "Ataxia Telangiectasia"
        },
        {
            "count": 2591,
            "id": "483",
            "name": "Spinal Muscular Atrophy"
        },
        {
            "count": 2552,
            "id": "586",
            "name": "Vascular Cognitive Impairment/Dementia"
        },
        {
            "count": 2414,
            "id": "459",
            "name": "Mucopolysaccharidoses (MPS)"
        },
        {
            "count": 2302,
            "id": "518",
            "name": "Eczema / Atopic Dermatitis"
        },
        {
            "count": 2277,
            "id": "530",
            "name": "Human Fetal Tissue"
        },
        {
            "count": 2230,
            "id": "470",
            "name": "Rett Syndrome"
        },
        {
            "count": 2098,
            "id": "474",
            "name": "Screening And Brief Intervention For Substance Abuse"
        },
        {
            "count": 2004,
            "id": "329",
            "name": "Topical Microbicides"
        },
        {
            "count": 1859,
            "id": "538",
            "name": "Lewy Body Dementia"
        },
        {
            "count": 1726,
            "id": "336",
            "name": "Climate Change"
        },
        {
            "count": 1633,
            "id": "449",
            "name": "Lead Poisoning"
        },
        {
            "count": 1524,
            "id": "456",
            "name": "Malaria Vaccine"
        },
        {
            "count": 1495,
            "id": "567",
            "name": "Stem Cell Research - Induced Pluripotent Stem Cell - Non-Human"
        },
        {
            "count": 1361,
            "id": "495",
            "name": "Arctic"
        },
        {
            "count": 1227,
            "id": "335",
            "name": "Batten Disease"
        },
        {
            "count": 812,
            "id": "528",
            "name": "Health Effects of Household Energy Combustion"
        },
        {
            "count": 792,
            "id": "355",
            "name": "Facioscapulohumeral Muscular Dystrophy"
        },
        {
            "count": 712,
            "id": "440",
            "name": "Global Warming Climate Change"
        },
        {
            "count": 494,
            "id": "576",
            "name": "Therapeutic Cannabinoid Research"
        },
        {
            "count": 427,
            "id": "404",
            "name": "Vulvodynia"
        },
        {
            "count": 423,
            "id": "581",
            "name": "Usher Syndrome"
        },
        {
            "count": 397,
            "id": "504",
            "name": "Cannabidiol Research"
        },
        {
            "count": 283,
            "id": "585",
            "name": "Valley Fever"
        },
        {
            "count": 215,
            "id": "555",
            "name": "Pick's Disease"
        },
        {
            "count": 208,
            "id": "551",
            "name": "Pelvic Inflammatory Disease"
        },
        {
            "count": 94,
            "id": "366",
            "name": "Hyperbaric Oxygen"
        },
        {
            "count": 82,
            "id": "515",
            "name": "Diethylstilbestrol (DES)"
        },
        {
            "count": 19,
            "id": "545",
            "name": "Paget's Disease"
        },
        {
            "count": 12,
            "id": "524",
            "name": "Gene Therapy Clinical Trials"
        },
        {
            "count": 9,
            "id": "572",
            "name": "Stem Cell Research - Umbilical Cord Blood/ Placenta - Non-Human"
        },
        {
            "count": 4,
            "id": "417",
            "name": "Hepatitis - A"
        },
        {
            "count": 2,
            "id": "319",
            "name": "Homicide and Legal Interventions"
        }
    ],
    "category_hrcs_rac": [
        {
            "count": 2396058,
            "id": "10201",
            "name": "2.1 Biological and endogenous factors"
        },
        {
            "count": 1214188,
            "id": "10101",
            "name": "1.1 Normal biological development and functioning"
        },
        {
            "count": 856699,
            "id": "10601",
            "name": "6.1 Pharmaceuticals"
        },
        {
            "count": 648809,
            "id": "10501",
            "name": "5.1 Pharmaceuticals"
        },
        {
            "count": 406362,
            "id": "10402",
            "name": "4.2 Evaluation of markers and technologies"
        },
        {
            "count": 371676,
            "id": "10202",
            "name": "2.2 Factors relating to physical environment"
        },
        {
            "count": 291583,
            "id": "10401",
            "name": "4.1 Discovery and preclinical testing of markers and technologies"
        },
        {
            "count": 249259,
            "id": "10604",
            "name": "6.4 Surgery"
        },
        {
            "count": 247021,
            "id": "10801",
            "name": "8.1 Organisation and delivery of services"
        },
        {
            "count": 246925,
            "id": "10701",
            "name": "7.1 Individual care needs"
        },
        {
            "count": 132635,
            "id": "10703",
            "name": "7.3 Management and decision making"
        },
        {
            "count": 124797,
            "id": "10203",
            "name": "2.3 Psychological, social and economic factors"
        },
        {
            "count": 118931,
            "id": "10502",
            "name": "5.2 Cellular and gene therapies"
        },
        {
            "count": 84768,
            "id": "10304",
            "name": "3.4 Vaccines"
        },
        {
            "count": 71809,
            "id": "10301",
            "name": "3.1 Primary prevention interventions to modify behaviours or promote well-being"
        },
        {
            "count": 66281,
            "id": "10204",
            "name": "2.4 Surveillance and distribution"
        },
        {
            "count": 63543,
            "id": "10303",
            "name": "3.3 Nutrition and chemoprevention"
        },
        {
            "count": 54768,
            "id": "10607",
            "name": "6.7 Physical"
        },
        {
            "count": 52397,
            "id": "10102",
            "name": "1.2 Psychological and socioeconomic processes"
        },
        {
            "count": 45279,
            "id": "10606",
            "name": "6.6 Psychological and behavioural"
        },
        {
            "count": 41337,
            "id": "10803",
            "name": "8.3 Policy, ethics and research governance"
        },
        {
            "count": 35695,
            "id": "10605",
            "name": "6.5 Radiotherapy"
        },
        {
            "count": 22275,
            "id": "10603",
            "name": "6.3 Medical devices"
        },
        {
            "count": 18798,
            "id": "10702",
            "name": "7.2 End of life care"
        },
        {
            "count": 16546,
            "id": "10804",
            "name": "8.4 Research design and methodologies"
        },
        {
            "count": 15451,
            "id": "10404",
            "name": "4.4 Population screening"
        },
        {
            "count": 13219,
            "id": "10602",
            "name": "6.2 Cellular and gene therapies"
        },
        {
            "count": 11869,
            "id": "10205",
            "name": "2.5 Research design and methodologies (aetiology)"
        },
        {
            "count": 8627,
            "id": "10802",
            "name": "8.2 Health and welfare economics"
        },
        {
            "count": 7994,
            "id": "10503",
            "name": "5.3 Medical devices"
        },
        {
            "count": 7557,
            "id": "10302",
            "name": "3.2 Interventions to alter physical and biological environmental risks"
        },
        {
            "count": 5320,
            "id": "10504",
            "name": "5.4 Surgery"
        },
        {
            "count": 4867,
            "id": "10509",
            "name": "5.9 Resources and infrastructure (development of treatments)"
        },
        {
            "count": 4492,
            "id": "10103",
            "name": "1.3 Chemical and physical sciences"
        },
        {
            "count": 4217,
            "id": "10505",
            "name": "5.5 Radiotherapy"
        },
        {
            "count": 3911,
            "id": "10206",
            "name": "2.6 Resources and infrastructure (aetiology)"
        },
        {
            "count": 3272,
            "id": "10105",
            "name": "1.5 Resources and infrastructure (underpinning)"
        },
        {
            "count": 2091,
            "id": "10104",
            "name": "1.4 Methodologies and measurements"
        },
        {
            "count": 1820,
            "id": "10305",
            "name": "3.5 Resources and infrastructure (prevention)"
        },
        {
            "count": 1150,
            "id": "10609",
            "name": "6.9 Resources and infrastructure (evaluation of treatments)"
        },
        {
            "count": 1018,
            "id": "10405",
            "name": "4.5 Resources and infrastructure (detection)"
        },
        {
            "count": 998,
            "id": "10506",
            "name": "5.6 Psychological and behavioural"
        },
        {
            "count": 983,
            "id": "10403",
            "name": "4.3 Influences and impact"
        },
        {
            "count": 448,
            "id": "10608",
            "name": "6.8 Complementary"
        },
        {
            "count": 374,
            "id": "10507",
            "name": "5.7 Physical"
        },
        {
            "count": 41,
            "id": "10805",
            "name": "8.5 Resources and infrastructure (health services)"
        },
        {
            "count": 2,
            "id": "10704",
            "name": "7.4 Resources and infrastructure (disease management)"
        }
    ],
    "category_hrcs_hc": [
        {
            "count": 1993062,
            "id": "911",
            "name": "Cancer"
        },
        {
            "count": 1631954,
            "id": "894",
            "name": "Cardiovascular"
        },
        {
            "count": 1456546,
            "id": "898",
            "name": "Infection"
        },
        {
            "count": 1073124,
            "id": "890",
            "name": "Generic Health Relevance"
        },
        {
            "count": 920146,
            "id": "897",
            "name": "Neurological"
        },
        {
            "count": 802034,
            "id": "905",
            "name": "Mental Health"
        },
        {
            "count": 610840,
            "id": "906",
            "name": "Metabolic and Endocrine"
        },
        {
            "count": 564794,
            "id": "903",
            "name": "Inflammatory and Immune System"
        },
        {
            "count": 553507,
            "id": "908",
            "name": "Reproductive Health and Childbirth"
        },
        {
            "count": 534806,
            "id": "899",
            "name": "Oral and Gastrointestinal"
        },
        {
            "count": 508497,
            "id": "900",
            "name": "Musculoskeletal"
        },
        {
            "count": 333324,
            "id": "907",
            "name": "Renal and Urogenital"
        },
        {
            "count": 329426,
            "id": "896",
            "name": "Respiratory"
        },
        {
            "count": 239268,
            "id": "909",
            "name": "Stroke"
        },
        {
            "count": 216912,
            "id": "895",
            "name": "Eye"
        },
        {
            "count": 158148,
            "id": "904",
            "name": "Injuries and Accidents"
        },
        {
            "count": 106998,
            "id": "891",
            "name": "Skin"
        },
        {
            "count": 60363,
            "id": "901",
            "name": "Congenital Disorders"
        },
        {
            "count": 56561,
            "id": "902",
            "name": "Ear"
        },
        {
            "count": 52556,
            "id": "892",
            "name": "Blood"
        },
        {
            "count": 11910,
            "id": "910",
            "name": "Other"
        }
    ],
    "category_hra": [
        {
            "count": 5839488,
            "id": "3900",
            "name": "Biomedical"
        },
        {
            "count": 4963490,
            "id": "3901",
            "name": "Clinical"
        },
        {
            "count": 1291542,
            "id": "3903",
            "name": "Population & Society"
        },
        {
            "count": 860794,
            "id": "3902",
            "name": "Health services & systems"
        }
    ],
    "category_bra": [
        {
            "count": 7217489,
            "id": "4001",
            "name": "Clinical Medicine and Science"
        },
        {
            "count": 4086209,
            "id": "4000",
            "name": "Basic Science"
        },
        {
            "count": 1165190,
            "id": "4003",
            "name": "Public Health"
        },
        {
            "count": 458177,
            "id": "4002",
            "name": "Health Services Research"
        }
    ],
    "category_for": [
        {
            "count": 25778454,
            "id": "2211",
            "name": "11 Medical and Health Sciences"
        },
        {
            "count": 10516989,
            "id": "2209",
            "name": "09 Engineering"
        },
        {
            "count": 9203750,
            "id": "3053",
            "name": "1103 Clinical Sciences"
        },
        {
            "count": 8103410,
            "id": "2206",
            "name": "06 Biological Sciences"
        },
        {
            "count": 7183561,
            "id": "2203",
            "name": "03 Chemical Sciences"
        },
        {
            "count": 5735050,
            "id": "2202",
            "name": "02 Physical Sciences"
        },
        {
            "count": 4460429,
            "id": "2201",
            "name": "01 Mathematical Sciences"
        },
        {
            "count": 4407004,
            "id": "2208",
            "name": "08 Information and Computing Sciences"
        },
        {
            "count": 4123089,
            "id": "3177",
            "name": "1117 Public Health and Health Services"
        },
        {
            "count": 3316164,
            "id": "2217",
            "name": "17 Psychology and Cognitive Sciences"
        },
        {
            "count": 3083175,
            "id": "2921",
            "name": "0912 Materials Engineering"
        },
        {
            "count": 2914925,
            "id": "2471",
            "name": "0306 Physical Chemistry (incl. Structural)"
        },
        {
            "count": 2859140,
            "id": "2581",
            "name": "0601 Biochemistry and Cell Biology"
        },
        {
            "count": 2682874,
            "id": "2216",
            "name": "16 Studies in Human Society"
        },
        {
            "count": 2457529,
            "id": "3468",
            "name": "1701 Psychology"
        },
        {
            "count": 2390866,
            "id": "3120",
            "name": "1109 Neurosciences"
        },
        {
            "count": 2313315,
            "id": "2746",
            "name": "0801 Artificial Intelligence and Image Processing"
        },
        {
            "count": 2263912,
            "id": "3048",
            "name": "1102 Cardiorespiratory Medicine and Haematology"
        },
        {
            "count": 1913932,
            "id": "2220",
            "name": "20 Language, Communication and Culture"
        },
        {
            "count": 1909259,
            "id": "3142",
            "name": "1112 Oncology and Carcinogenesis"
        },
        {
            "count": 1851145,
            "id": "2207",
            "name": "07 Agricultural and Veterinary Sciences"
        },
        {
            "count": 1810707,
            "id": "2204",
            "name": "04 Earth Sciences"
        },
        {
            "count": 1695157,
            "id": "2221",
            "name": "21 History and Archaeology"
        },
        {
            "count": 1686771,
            "id": "2620",
            "name": "0604 Genetics"
        },
        {
            "count": 1685238,
            "id": "2330",
            "name": "0101 Pure Mathematics"
        },
        {
            "count": 1610939,
            "id": "2210",
            "name": "10 Technology"
        },
        {
            "count": 1482270,
            "id": "2867",
            "name": "0906 Electrical and Electronic Engineering"
        },
        {
            "count": 1475153,
            "id": "2215",
            "name": "15 Commerce, Management, Tourism and Services"
        },
        {
            "count": 1454065,
            "id": "2421",
            "name": "0299 Other Physical Sciences"
        },
        {
            "count": 1451514,
            "id": "2213",
            "name": "13 Education"
        },
        {
            "count": 1401999,
            "id": "3675",
            "name": "2103 Historical Studies"
        },
        {
            "count": 1391768,
            "id": "3158",
            "name": "1114 Paediatrics and Reproductive Medicine"
        },
        {
            "count": 1390528,
            "id": "2214",
            "name": "14 Economics"
        },
        {
            "count": 1313414,
            "id": "2389",
            "name": "0202 Atomic, Molecular, Nuclear, Particle and Plasma Physics"
        },
        {
            "count": 1198971,
            "id": "2222",
            "name": "22 Philosophy and Religious Studies"
        },
        {
            "count": 1181174,
            "id": "2205",
            "name": "05 Environmental Sciences"
        },
        {
            "count": 1175049,
            "id": "3103",
            "name": "1107 Immunology"
        },
        {
            "count": 1089260,
            "id": "2790",
            "name": "0806 Information Systems"
        },
        {
            "count": 1050254,
            "id": "3164",
            "name": "1115 Pharmacology and Pharmaceutical Sciences"
        },
        {
            "count": 1020219,
            "id": "3292",
            "name": "1402 Applied Economics"
        },
        {
            "count": 1016796,
            "id": "2344",
            "name": "0102 Applied Mathematics"
        },
        {
            "count": 996582,
            "id": "3626",
            "name": "2005 Literary Studies"
        },
        {
            "count": 976222,
            "id": "2953",
            "name": "0915 Interdisciplinary Engineering"
        },
        {
            "count": 946002,
            "id": "2648",
            "name": "0607 Plant Biology"
        },
        {
            "count": 937408,
            "id": "3001",
            "name": "1005 Communications Technologies"
        },
        {
            "count": 906219,
            "id": "2844",
            "name": "0904 Chemical Engineering"
        },
        {
            "count": 882031,
            "id": "2377",
            "name": "0201 Astronomical and Space Sciences"
        },
        {
            "count": 880822,
            "id": "2856",
            "name": "0905 Civil Engineering"
        },
        {
            "count": 873634,
            "id": "2464",
            "name": "0305 Organic Chemistry"
        },
        {
            "count": 856978,
            "id": "2597",
            "name": "0602 Ecology"
        },
        {
            "count": 817545,
            "id": "3342",
            "name": "1503 Business and Management"
        },
        {
            "count": 809779,
            "id": "3268",
            "name": "1303 Specialist Studies In Education"
        },
        {
            "count": 788132,
            "id": "2438",
            "name": "0302 Inorganic Chemistry"
        },
        {
            "count": 777625,
            "id": "3484",
            "name": "1702 Cognitive Sciences"
        },
        {
            "count": 767507,
            "id": "3114",
            "name": "1108 Medical Microbiology"
        },
        {
            "count": 759134,
            "id": "2428",
            "name": "0301 Analytical Chemistry"
        },
        {
            "count": 752809,
            "id": "3432",
            "name": "1606 Political Science"
        },
        {
            "count": 752183,
            "id": "2509",
            "name": "0403 Geology"
        },
        {
            "count": 732917,
            "id": "2358",
            "name": "0104 Statistics"
        },
        {
            "count": 717304,
            "id": "3448",
            "name": "1608 Sociology"
        },
        {
            "count": 717147,
            "id": "3128",
            "name": "1110 Nursing"
        },
        {
            "count": 711579,
            "id": "2933",
            "name": "0913 Mechanical Engineering"
        },
        {
            "count": 680860,
            "id": "2486",
            "name": "0399 Other Chemical Sciences"
        },
        {
            "count": 651170,
            "id": "2447",
            "name": "0303 Macromolecular and Materials Chemistry"
        },
        {
            "count": 602973,
            "id": "2837",
            "name": "0903 Biomedical Engineering"
        },
        {
            "count": 587365,
            "id": "2634",
            "name": "0605 Microbiology"
        },
        {
            "count": 581263,
            "id": "2218",
            "name": "18 Law and Legal Studies"
        },
        {
            "count": 577386,
            "id": "2655",
            "name": "0608 Zoology"
        },
        {
            "count": 564949,
            "id": "3494",
            "name": "1801 Law"
        },
        {
            "count": 563884,
            "id": "3253",
            "name": "1302 Curriculum and Pedagogy"
        },
        {
            "count": 529033,
            "id": "3714",
            "name": "2203 Philosophy"
        },
        {
            "count": 495600,
            "id": "2558",
            "name": "0502 Environmental Science and Management"
        },
        {
            "count": 488081,
            "id": "3416",
            "name": "1605 Policy and Administration"
        },
        {
            "count": 483065,
            "id": "3616",
            "name": "2004 Linguistics"
        },
        {
            "count": 476079,
            "id": "2766",
            "name": "0803 Computer Software"
        },
        {
            "count": 467726,
            "id": "2456",
            "name": "0304 Medicinal and Biomolecular Chemistry"
        },
        {
            "count": 461587,
            "id": "2539",
            "name": "0406 Physical Geography and Environmental Geoscience"
        },
        {
            "count": 448145,
            "id": "2878",
            "name": "0907 Environmental Engineering"
        },
        {
            "count": 447376,
            "id": "2401",
            "name": "0204 Condensed Matter Physics"
        },
        {
            "count": 446883,
            "id": "2409",
            "name": "0205 Optical Physics"
        },
        {
            "count": 428069,
            "id": "2727",
            "name": "0707 Veterinary Sciences"
        },
        {
            "count": 427348,
            "id": "2219",
            "name": "19 Studies in Creative Arts and Writing"
        },
        {
            "count": 425377,
            "id": "2353",
            "name": "0103 Numerical and Computational Mathematics"
        },
        {
            "count": 421440,
            "id": "3153",
            "name": "1113 Ophthalmology and Optometry"
        },
        {
            "count": 418060,
            "id": "3735",
            "name": "2204 Religion and Religious Studies"
        },
        {
            "count": 415592,
            "id": "2944",
            "name": "0914 Resources Engineering and Extractive Metallurgy"
        },
        {
            "count": 412950,
            "id": "3577",
            "name": "2002 Cultural Studies"
        },
        {
            "count": 399440,
            "id": "2759",
            "name": "0802 Computation Theory and Mathematics"
        },
        {
            "count": 395771,
            "id": "3097",
            "name": "1106 Human Movement and Sports Science"
        },
        {
            "count": 383243,
            "id": "2690",
            "name": "0703 Crop and Pasture Production"
        },
        {
            "count": 381415,
            "id": "3172",
            "name": "1116 Medical Physiology"
        },
        {
            "count": 371724,
            "id": "3086",
            "name": "1105 Dentistry"
        },
        {
            "count": 366798,
            "id": "2212",
            "name": "12 Built Environment and Design"
        },
        {
            "count": 352120,
            "id": "2607",
            "name": "0603 Evolutionary Biology"
        },
        {
            "count": 346160,
            "id": "2415",
            "name": "0206 Quantum Physics"
        },
        {
            "count": 329082,
            "id": "2883",
            "name": "0908 Food Sciences"
        },
        {
            "count": 303990,
            "id": "2480",
            "name": "0307 Theoretical and Computational Chemistry"
        },
        {
            "count": 286009,
            "id": "3702",
            "name": "2202 History and Philosophy of Specific Fields"
        },
        {
            "count": 274236,
            "id": "2493",
            "name": "0401 Atmospheric Sciences"
        },
        {
            "count": 270570,
            "id": "3410",
            "name": "1604 Human Geography"
        },
        {
            "count": 253142,
            "id": "2503",
            "name": "0402 Geochemistry"
        },
        {
            "count": 252038,
            "id": "2571",
            "name": "0503 Soil Sciences"
        },
        {
            "count": 235565,
            "id": "3364",
            "name": "1505 Marketing"
        },
        {
            "count": 232293,
            "id": "2899",
            "name": "0910 Manufacturing Engineering"
        },
        {
            "count": 230077,
            "id": "3657",
            "name": "2101 Archaeology"
        },
        {
            "count": 226473,
            "id": "3389",
            "name": "1601 Anthropology"
        },
        {
            "count": 223619,
            "id": "3549",
            "name": "1904 Performing Arts and Creative Writing"
        },
        {
            "count": 222831,
            "id": "3335",
            "name": "1502 Banking, Finance and Investment"
        },
        {
            "count": 217407,
            "id": "2891",
            "name": "0909 Geomatic Engineering"
        },
        {
            "count": 209186,
            "id": "3039",
            "name": "1101 Medical Biochemistry and Metabolomics"
        },
        {
            "count": 207478,
            "id": "3313",
            "name": "1403 Econometrics"
        },
        {
            "count": 205583,
            "id": "2777",
            "name": "0804 Data Format"
        },
        {
            "count": 203662,
            "id": "2681",
            "name": "0702 Animal Production"
        },
        {
            "count": 203186,
            "id": "2525",
            "name": "0404 Geophysics"
        },
        {
            "count": 199918,
            "id": "3136",
            "name": "1111 Nutrition and Dietetics"
        },
        {
            "count": 194217,
            "id": "3460",
            "name": "1699 Other Studies In Human Society"
        },
        {
            "count": 192246,
            "id": "3021",
            "name": "1007 Nanotechnology"
        },
        {
            "count": 185508,
            "id": "2534",
            "name": "0405 Oceanography"
        },
        {
            "count": 161813,
            "id": "2395",
            "name": "0203 Classical Physics"
        },
        {
            "count": 156708,
            "id": "2642",
            "name": "0606 Physiology"
        },
        {
            "count": 155471,
            "id": "3591",
            "name": "2003 Language Studies"
        },
        {
            "count": 153128,
            "id": "2820",
            "name": "0901 Aerospace Engineering"
        },
        {
            "count": 149166,
            "id": "3443",
            "name": "1607 Social Work"
        },
        {
            "count": 148579,
            "id": "2720",
            "name": "0706 Horticultural Production"
        },
        {
            "count": 146221,
            "id": "2806",
            "name": "0807 Library and Information Studies"
        },
        {
            "count": 146018,
            "id": "2708",
            "name": "0705 Forestry Sciences"
        },
        {
            "count": 133660,
            "id": "2552",
            "name": "0501 Ecological Applications"
        },
        {
            "count": 130749,
            "id": "3230",
            "name": "1205 Urban and Regional Planning"
        },
        {
            "count": 130367,
            "id": "3197",
            "name": "1199 Other Medical and Health Sciences"
        },
        {
            "count": 127476,
            "id": "3243",
            "name": "1301 Education Systems"
        },
        {
            "count": 127292,
            "id": "2700",
            "name": "0704 Fisheries Sciences"
        },
        {
            "count": 121308,
            "id": "2366",
            "name": "0105 Mathematical Physics"
        },
        {
            "count": 116918,
            "id": "3403",
            "name": "1603 Demography"
        },
        {
            "count": 112406,
            "id": "3395",
            "name": "1602 Criminology"
        },
        {
            "count": 110597,
            "id": "3570",
            "name": "2001 Communication and Media Studies"
        },
        {
            "count": 106196,
            "id": "2913",
            "name": "0911 Maritime Engineering"
        },
        {
            "count": 100125,
            "id": "3373",
            "name": "1506 Tourism"
        },
        {
            "count": 89877,
            "id": "3326",
            "name": "1501 Accounting, Auditing and Accountability"
        },
        {
            "count": 86730,
            "id": "3286",
            "name": "1401 Economic Theory"
        },
        {
            "count": 84952,
            "id": "3209",
            "name": "1202 Building"
        },
        {
            "count": 82032,
            "id": "3079",
            "name": "1104 Complementary and Alternative Medicine"
        },
        {
            "count": 77019,
            "id": "2783",
            "name": "0805 Distributed Computing"
        },
        {
            "count": 72399,
            "id": "2995",
            "name": "1004 Medical Biotechnology"
        },
        {
            "count": 71811,
            "id": "2671",
            "name": "0701 Agriculture, Land and Farm Management"
        },
        {
            "count": 69757,
            "id": "3693",
            "name": "2201 Applied Ethics"
        },
        {
            "count": 69686,
            "id": "3200",
            "name": "1201 Architecture"
        },
        {
            "count": 69113,
            "id": "3358",
            "name": "1504 Commercial Services"
        },
        {
            "count": 68713,
            "id": "3537",
            "name": "1902 Film, Television and Digital Media"
        },
        {
            "count": 67234,
            "id": "3531",
            "name": "1901 Art Theory and Criticism"
        },
        {
            "count": 63182,
            "id": "3214",
            "name": "1203 Design Practice and Management"
        },
        {
            "count": 56894,
            "id": "2987",
            "name": "1003 Industrial Biotechnology"
        },
        {
            "count": 56543,
            "id": "3669",
            "name": "2102 Curatorial and Related Studies"
        },
        {
            "count": 54928,
            "id": "2666",
            "name": "0699 Other Biological Sciences"
        },
        {
            "count": 53602,
            "id": "3381",
            "name": "1507 Transportation and Freight Services"
        },
        {
            "count": 47276,
            "id": "3544",
            "name": "1903 Journalism and Professional Writing"
        },
        {
            "count": 44855,
            "id": "2817",
            "name": "0899 Other Information and Computing Sciences"
        },
        {
            "count": 36989,
            "id": "3013",
            "name": "1006 Computer Hardware"
        },
        {
            "count": 35739,
            "id": "3561",
            "name": "1905 Visual Arts and Crafts"
        },
        {
            "count": 34187,
            "id": "3320",
            "name": "1499 Other Economics"
        },
        {
            "count": 26503,
            "id": "2830",
            "name": "0902 Automotive Engineering"
        },
        {
            "count": 19415,
            "id": "2963",
            "name": "0999 Other Engineering"
        },
        {
            "count": 18971,
            "id": "3283",
            "name": "1399 Other Education"
        },
        {
            "count": 18317,
            "id": "2968",
            "name": "1001 Agricultural Biotechnology"
        },
        {
            "count": 16916,
            "id": "2741",
            "name": "0799 Other Agricultural and Veterinary Sciences"
        },
        {
            "count": 16750,
            "id": "3036",
            "name": "1099 Other Technology"
        },
        {
            "count": 15461,
            "id": "2979",
            "name": "1002 Environmental Biotechnology"
        },
        {
            "count": 14837,
            "id": "3690",
            "name": "2199 Other History and Archaeology"
        },
        {
            "count": 7876,
            "id": "3654",
            "name": "2099 Other Language, Communication and Culture"
        },
        {
            "count": 5693,
            "id": "2549",
            "name": "0499 Other Earth Sciences"
        },
        {
            "count": 3186,
            "id": "3491",
            "name": "1799 Other Psychology and Cognitive Sciences"
        },
        {
            "count": 2877,
            "id": "3528",
            "name": "1899 Other Law and Legal Studies"
        },
        {
            "count": 2482,
            "id": "3567",
            "name": "1999 Other Studies In Creative Arts and Writing"
        },
        {
            "count": 730,
            "id": "3744",
            "name": "2299 Other Philosophy and Religious Studies"
        },
        {
            "count": 644,
            "id": "2578",
            "name": "0599 Other Environmental Sciences"
        },
        {
            "count": 237,
            "id": "3240",
            "name": "1299 Other Built Environment and Design"
        }
    ]
}



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit tests for Dimcli - DfConverter
"""

from __future__ import print_function

import unittest, os, sys, click
import configparser

from .. import *
from ..core.auth import USER_CONFIG_FILE_PATH
from ..core.utils import dimensions_url
from ..shortcuts import *

from ..core.converters import *


class TestRenderer(unittest.TestCase):

    """
    Tests  
    """

    click.secho("**TESTS**", fg="red")
    login(instance="live")
    d = Dsl()

    def test_001(self):
        click.secho("\nTEST 001: DfConverter.", fg="green")
        # ----
        q = "search publications for \"mercedes\" return publications"
        click.secho(q, fg="green")
        res = dslquery(q).as_dataframe()

        r = DfConverter(res)
        r.do()
        print(r.df)




if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from prompt_toolkit.history import FileHistory

from ..core.dsl_grammar import *

#
#
# HISTORY
#
#


class SelectiveFileHistory(FileHistory):
    """
    :class:`.SelectiveFileHistory` class that extends history but stores only queries 
     - strings starting with 'search' 
    NOTE This approach can be refined in the future
    """

    def __init__(self, filename):
        self.filename = filename
        super(SelectiveFileHistory, self).__init__(filename)

    def append_string(self, string):
        " Add string to the history only if it is a valid DSL query"
        l = G.allowed_starts_dsl_query()
        for x in l:
            if string.startswith(x):
                self._loaded_strings.append(string)
                self.store_string(string)
                return


#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# LEXER AND SYNTAX HIGHLIGHTING
#
# https://python-prompt-toolkit.readthedocs.io/en/master/pages/reference.html#prompt_toolkit.lexers.Lexer
#
#

from prompt_toolkit.lexers import Lexer
from ..core.dsl_grammar import *
from ..core.utils import *



# ====
#
allowed_starts = G.allowed_starts()
lang = split_multi_words(G.lang())
#
sources = G.sources()
entities = G.entities()
#
fields = list_flatten([G.fields_for_source(source) for source in sources])
facets = list_flatten([G.facets_for_source(source) for source in sources])
fieldsets = list_flatten([G.fieldsets_for_source(source) for source in sources])
metrics = list_flatten([G.metrics_for_source(source) for source in sources])
search_fields = list_flatten([G.search_fields_for_source(source) for source in sources])
#
# ====

class BasicLexer(Lexer):
    """
    note: lexer is single word based, so multi word DSL operands need to split first 
    """
    def lex_document(self, document):

        def get_class(w):
            "color classes for main objects - color-safe"
            if w in lang:
                return "bold"
            elif w in sources + entities:
                return "bold"
            elif w in fields:
                return "italic"
            elif w in facets:
                return "italic" 
            elif w in metrics:
                return "italic" 
            elif w in fieldsets + search_fields:
                return "italic" 
            elif w in allowed_starts:
                return "bold"
            else:
                return "nobold"

        def previous_get_class(w):
            "DEPRECATED DUE TO UNREADABILITY ON SOME TERMINALS - color classes for main objects"
            if w in lang:
                return "green"
            elif w in sources + entities:
                return "blue bold"
            elif w in fields:
                return "blue"
            elif w in facets:
                return "blue" 
            elif w in metrics:
                return "brown" 
            elif w in fieldsets + search_fields:
                return "black" 
            elif w in allowed_starts:
                return "red"
            else:
                return "black"

        def _spot_strings_bits(data):
            "note: this is run after the whole sentence has been parsed and marked up, so to be able to spot multi-word strings"
            STRING_COLOR = "red"
            is_string_flag = False
            for x in data:
                if is_string_flag:
                    x[0] = STRING_COLOR
                    if x[1].strip()[-1] == '"':
                        is_string_flag = False
                elif x[1].strip()[0] == '"' and x[1].strip()[-1] == '"':
                    x[0] = STRING_COLOR
                elif x[1].strip()[0] == '"':
                    x[0] = STRING_COLOR
                    is_string_flag = True
            return data

        def get_line(lineno):
            "NOTE: this gets called at each single key press, i.e. line is re-rendered all the time"
            data = [[get_class(w), w + " "]
                    for w in document.lines[lineno].split()]
            # add on to mark up multi-word strings post-process
            data = _spot_strings_bits(data)
            return data

        return get_line



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit tests for Dimcli  - logins
"""

from __future__ import print_function

import unittest, os, sys, click
import configparser

from .. import *
from ..core.auth import USER_CONFIG_FILE_PATH
from ..core.utils import dimensions_url
from ..shortcuts import *


class TestOne(unittest.TestCase):

    """
    Tests  
    """

    click.secho("**TESTS**", fg="red")

    def test_001(self):
        click.secho("\nTEST 001: checking that grammar data is valid.", fg="green")
        # ----
        g = G
        print(g)
        print("*STARTS*", g.allowed_starts())
        print("*LANG*", g.lang())
        print("*SOURCES*", g.sources())
        print("*ENTITIES*", g.entities())
        print("*CATEGORIES*", g.categories())
        for x in g.sources():
            click.secho("============\n" + x + "\n============",  fg="magenta")
            print(g.url_for_source(x))
            for y in g.fields_for_source(x):
                print("...", y, " => ", str(g.desc_for_source_field(x, y)), "**facet?**", str(g.entity_type_for_source_facet(x, y)), str(g.fields_for_entity_from_source_facet(x, y)))

        for x in g.entities():
            click.secho("============\n" + x + "\n============",  fg="magenta")
            for y in g.fields_for_entity(x):
                print("...",  y, " => ", str(g.desc_for_entity_field(x, y)))
        for x in g.categories():
            click.secho("============\n" + x + "\n============",  fg="magenta")
            for y in g.categories(x):
                print("...",  y)


    def test_002(self):
        click.secho("\nTEST 002: dimensions URL generation.", fg="green")
        # ----
        click.secho("""....Testing dimensions_url.... dimensions_url("01", "stff")""",  fg="magenta")
        try:
            print(dimensions_url("01", "stff"))
        except Exception as e:
            print("ERROR:",  e)
        click.secho("""....Testing dimensions_url.... dimensions_url("01", "publications")""",  fg="magenta")
        print(dimensions_url("01", "publications"))
        click.secho("""....Testing dimensions_url....   dimensions_url("pub.1122527319")""",  fg="magenta")
        print(dimensions_url("pub.1122527319"))
        # grants https://app.dimensions.ai/details/grant/grant.8587603
        click.secho("""....Testing dimensions_url....   dimensions_url("01", "grants")""",  fg="magenta")
        print(dimensions_url("01", "grants"))
        click.secho("""....Testing dimensions_url....   dimensions_url("grant.8587603")""",  fg="magenta")
        print(dimensions_url("grant.8587603"))
        # patents https://app.dimensions.ai/details/patent/IN-293637-B
        click.secho("""....Testing dimensions_url....   dimensions_url("IN-293637-B", "patents")""",  fg="magenta")
        print(dimensions_url("IN-293637-B", "patents"))
        click.secho("""....Testing dimensions_url....   dimensions_url("IN-293637-B") [FAILS!]""",  fg="magenta")
        print(dimensions_url("IN-293637-B"))
        # researchers https://app.dimensions.ai/discover/patent?and_facet_researcher=ur.01117537642.02
        click.secho("""....Testing dimensions_url....   dimensions_url("ur.01117537642.02")""",  fg="magenta")
        print(dimensions_url("ur.01117537642.02"))
        click.secho("""....Testing dimensions_url....   dimensions_url("ur.01117537642.02", "researchers")""",  fg="magenta")
        print(dimensions_url("ur.01117537642.02", "researchers"))
        # organizations https://app.dimensions.ai/discover/patent?and_facet_research_org=grid.2515.3
        click.secho("""....Testing dimensions_url....   dimensions_url("grid.2515.3", "organizations")""",  fg="magenta")
        print(dimensions_url("grid.2515.3", "organizations"))
        click.secho("""....Testing dimensions_url....   dimensions_url("grid.2515.3")""",  fg="magenta")
        print(dimensions_url("grid.2515.3"))
        # ----
        click.secho("Completed test succesfully", fg="green")


    def test_003(self):
        click.secho("\nTEST 003: Language utils: DSL escape.", fg="green")
        login()
        d = Dsl()
        # ----
        covid_q = '"2019-nCoV" OR "COVID-19" OR "SARS-CoV-2" OR (("coronavirus"  OR "corona virus") AND (Wuhan OR China))'
        q = f"""search publications 
                in full_data for "{dsl_escape(covid_q)}"  
            return publications[id+doi+pmid+pmcid+title+journal+publisher+mesh_terms+date+year+volume+issue+pages+open_access_categories+type+authors+research_orgs+funders+supporting_grant_ids+times_cited+altmetric+linkout] limit 1"""
        click.secho(q, fg="green")
        print("Success: ", bool(dslquery(q).count_total))
        # ----
        specialchar_q = 'Solar cells: a new technology? (some examples)'
        q = f"""search publications 
                in full_data for "{dsl_escape(specialchar_q, all=True)}"  
            return publications[id+doi+title] limit 1"""
        click.secho(q, fg="green")
        print("Success: ", bool(dslquery(q).count_total))
        # ----
        click.secho("Completed test succesfully", fg="green")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from ..VERSION import VERSION
import time
# from json2html import *


def html_template_interactive(query, formatted_json, api_endpoint):
    """
    This version just uses https://highlightjs.org/ to colorize the json code
    """

    this_time = time.strftime("%Y-%m-%d  at %H:%M:%S")

    s = """
    <html>
    <head>
        <title>Dimensions API Query Output</title>
        <meta http-equiv="content-type" content="text/html; charset=UTF-8">
        <style>
            .query {
                font-size: 20px;
                padding: 50px;
                margin-bottom: 30px;
                color: #f0f0f0;
                background: #2a6683;
                font-family: monospace;
            }

            .export-created {
                font-size: 14px;
                font-weight: normal;
                color: gray;
            }


        </style>
        <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.14.2/styles/default.min.css" />
        <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.14.2/highlight.min.js"></script>
        <script>hljs.initHighlightingOnLoad();</script>
    </head>
    <div>
        <div id="title">
            <h1>Dimensions API Export
                <span class="export-created">Created on: %s using the API endpoint: <a href="%s" target="_blank">%s</a></span>
            </h1>
            <hr>
        </div>      
        <div id="query">
            <h2>DSL Query:</h2>
            <p class="query">%s</p></API>
            
        </div>
        <hr>
        <h3>JSON Results:</h3>
        <a name="json">
        <div id="code"></a><pre><code>%s</code></pre>
        </div> 
        <div id="footer">
            <hr>
            <p>Generated with <a href="https://github.com/digital-science/dimcli">Dimcli</a> %s | See also: <a href="https://docs.dimensions.ai/dsl/">https://docs.dimensions.ai/dsl/</a></p>
        </div>
    </body>
    </html>
    
    """ % ( this_time, api_endpoint, api_endpoint,
            query.replace("search", "<b><i>search</i></b>").replace("return", "<b><i>return</i></b>"), 
            formatted_json , 
            VERSION)
    return s





import configparser
import requests
import os.path
import os
import time
import json
import click
import IPython.display
from itertools import islice
import urllib.parse

import pandas as pd

from .auth import do_global_login, get_connection, refresh_login
from .utils import *
from .walkup import *
from .dsl_grammar import G
from .dataframe_factory import DfFactory



class Dsl():
    """
    Object for abstracting common interaction steps with the Dimensions API. 
    Most often you just want to instantiate and query() - yeah!

    >>> import dimcli
    # if you have already set up the credentials file (see above), no need to pass log in details
    >>> dimcli.login()
    # otherwise you can authenticate by passing your login details as arguments
    >>> dimcli.login(user="mary.poppins", password="chimneysweeper")
    # instantiate the query object
    >>> dsl = dimcli.Dsl()
    # queries always return a DslDataset object (subclassing IPython.display.JSON)
    >>> dsl.query("search grants for \"malaria\" return researchers")
    >>> <dimcli.dimensions.DslDataset object>
    # use the .json method to get the JSON
    >>> dsl.query("search grants for \"malaria\" return researchers").json
    >>> {'researchers': [{'id': 'ur.01332073522.49',
            'count': 75,
            'last_name': 'White',
            'first_name': 'Nicholas J'},
        "... JSON data continues ... "

    """
    def __init__(self, show_results=False, verbose=True):
        # print(os.getcwd())
        self._show_results = show_results
        self._verbose = verbose
        self._url = None
        self._headers = None
        self._CONNECTION = get_connection()

        if self._CONNECTION['token']:
            # if already logged in, reuse connection          
            self._url = self._CONNECTION['url']
            self._headers = {'Authorization': "JWT " + self._CONNECTION['token']}
        else:
            self._print_please_login()

    @property
    def is_logged_in(self):
        if self._url and  self._headers: return True
        else: return False

    def _print_please_login(self):
        print("Warning: you are not logged in. Please use `dimcli.login(username, password)` before querying.")

    def query(self, q, show_results=None, retry=0, verbose=None):
        """
        Execute a DSL query.
        By default it doesn't show results, but it uses the iPython rich widgets for it, optimized for Jupyter Notebooks.
        """
        if not self.is_logged_in:
            self._print_please_login()
            return False

        if verbose == None:
            verbose = self._verbose

        #   Execute DSL query.
        response = requests.post(
            '{}/api/dsl.json'.format(self._url), data=q.encode(), headers=self._headers)
        if response.status_code == 429:  
            # Too Many Requests
            print(
                'Too Many Requests for the Server. Sleeping for 30 seconds and then retrying.'
            )
            time.sleep(30)
            return self.query(q, show_results, retry, verbose)
        elif response.status_code == 403:  
            # Forbidden:
            print('Login token expired. Logging in again.')
            refresh_login()
            self._CONNECTION = get_connection()
            self._url = self._CONNECTION['url']
            self._headers = {'Authorization': "JWT " + self._CONNECTION['token']}
            return self.query(q, show_results, retry, verbose)
        elif response.status_code in [200, 400, 500]:  
            ###  
            # OK or Error Info :-)
            ###
            try:
                res_json = response.json()
            except:
                print('Unexpected error. JSON could not be parsed.')
                return response
            result = DslDataset(res_json)
            if verbose: print_json_stats(result, q)
            print_json_errors(result) # ALWAYS print errors
            if verbose: print_json_warnings(result)
            if show_results or (show_results is None and self._show_results):
                IPython.display.display(result)
            return result
        else:
            if retry > 0:
                print('Retrying in 30 secs')
                time.sleep(30)
                return self.query(
                    q,
                    show_results,
                    retry - 1, 
                    verbose)
            else:
                response.raise_for_status()



    def query_iterative(self, q, show_results=None, limit=1000, skip=0, pause=1.5, force=False, verbose=None, tot_count_prev_query=0):       
        """Runs a DSL query iteratively, by automatically turning it into a loop with limit/skip operators until all the results available have been extracted.
        
        Args:
            q (str): The DSL query.
            show_results (bool): Determines whether the final results are rendered via the iPython display widget (for Jupyter notebooks).
            limit (int): How many records to extract per iteration. Defaults to 1000.
            skip (int): Offset for first iteration. Defaults to 0. After the first iteration, this value is calculated dynamically.
            pause (float): How much time to pause after each iterarion, expressed in seconds. Defaults to 1.5. Note: each iteration gets timed, so the pause time is used only when the query time is more than 2s. 
            verbose (bool): Verbose mode.

        
        Returns:
            DslDataset -- query results collated within a single object 
        """
        if not self.is_logged_in:
            self._print_please_login()
            return False

        if verbose == None:
            verbose = self._verbose

        if q.split().count('return') != 1:
            raise Exception("Iterative queries support only 1 return statement")
        if line_has_limit_or_skip(q):
            raise Exception("Iterative queries should not contain the keywords `limit` or `skip`")
        sourcetype = line_search_return(q)  
        if not (sourcetype in G.sources()):
            raise Exception("Iterative queries can return only one of the Dimensions sources: %s" % ", ".join([s for s in G.sources()])) 
        #
        # ensure we stop the loop at 50k **
        #
        MAXLIMIT = 50000
        flag_last_round = False
        if skip + limit >= MAXLIMIT:
            flag_last_round = True
            if skip + limit > MAXLIMIT:
                limit = MAXLIMIT - skip


        if not tot_count_prev_query:
            # first iteration
            if verbose: print(f"{limit+skip} / ...")
            
        output, flag_force = [], False
        q2 = q + " limit %d skip %d" % (limit, skip)
        
        start = time.time()
        res = self.query(q2, show_results=False, retry=0, verbose=False)
        end = time.time()
        # print(end - start)
        if (end - start) < 2:
            # print("sleeping")
            time.sleep(pause)

        if res['errors'] and not force:
            print(f"\n>>>[Dimcli tip] An error occurred with the batch '{skip}-{limit+skip}'. Consider using the 'limit' argument to retrieve fewer records per iteration, or use 'force=True' to ignore errors and continue the extraction.")
            return res
        elif res['errors'] and force:
            print(f"\n>>>[Dimcli log] An error occurred with the batch '{skip}-{limit+skip}'. Skipping this batch and continuing iteration.. ")
            flag_force = True

        # RECURSION 

        try:
            tot = int(res['stats']['total_count'])
        except:
            tot =  tot_count_prev_query # when force=True, we have no current query stats

        new_skip = skip+limit
        if tot > 0 and new_skip > tot:
            new_skip = tot
        if verbose and tot:  # if not first iteration
            print(f"{new_skip} / {tot}")

        if flag_force:
            output = self.query_iterative(q, show_results, limit, new_skip, pause, force, verbose, tot_count_prev_query)                    

        elif len(res[sourcetype]) == limit and not flag_last_round:
            output = res[sourcetype] + self.query_iterative(q, show_results, limit, new_skip, pause, force, verbose, tot)

        else:
            output = res[sourcetype]

        # FINALLY 
        #
        # if recursion is complete (we are at top level, hence skip=0) 
        #   build the DslDataset obj
        # else 
        #   just return current iteration results 
        #
        if skip == 0: 
            response_simulation = {
                "_stats": {
                    "total_count": tot or len(output)  # fallback..
                    },
                sourcetype: output
            }
            result = DslDataset(response_simulation)
            if show_results or (show_results is None and self._show_results):
                IPython.display.display(result)
            if verbose: print(f"===\nRecords extracted: {len(output)}")
            return result
        else:
            return output


    def __repr__(self):
        return f"<dimcli.Dsl #{id(self)}. API endpoint: {self._url}>"



        

class DslDataset(IPython.display.JSON):
    """
    Wrapper for JSON results from DSL

    >>> res = dsl.query("search publications return publications")
    >>> res.data # => shows the underlying JSON data
    >>> res.json # => same 

    # Magic methods: 

    >>> res['publications'] # => the dict section
    >>> res['xxx'] # => false, not found
    >>> res['stats'] # => the _stats dict

    """

    # class methods to build DslDataset object from raw data (obtained not from a query eg from multiple queries concatenated)
    # these allow to then take advantage of other functionalities in DslDataset objects eg dataframes etc...
    @classmethod
    def from_publications_list(cls, data):
        return cls.from_any_list(data, "publications")
    @classmethod
    def from_grants_list(cls, data):
        return cls.from_any_list(data, "grants")
    @classmethod
    def from_researchers_list(cls, data):
        return cls.from_any_list(data, "researchers")
    @classmethod
    def from_clinical_trials_list(cls, data):
        return cls.from_any_list(data, "clinical_trials")
    @classmethod
    def from_patents_list(cls, data):
        return cls.from_any_list(data, "patents")
    @classmethod
    def from_policy_documents_list(cls, data):
        return cls.from_any_list(data, "policy_documents")
    @classmethod
    def from_organizations_list(cls, data):
        return cls.from_any_list(data, "organizations")
    @classmethod
    def from_any_list(cls, data, source_type):
        "Generic method used by all the ones above"
        if type(data) == list:
            return cls({source_type : data, '_stats' : {'total_count' : len(data)}})
        elif type(data) == pd.DataFrame:
            jsondata = json.loads(data.to_json(orient="records"))
            return cls({source_type : jsondata, '_stats' : {'total_count' : len(jsondata)}})
        else:
            raise ValueError('Invalid data format. Must be either a dict list, or a pandas dataframe')



    def __init__(self, data):
        IPython.display.JSON.__init__(self, data)
        self.json = self.data
        self.errors = None
        for k in self.json.keys(): # add result dict keys as attributes dynamically
            if k == "_stats":
                setattr(self, "stats", self.json[k])
            else:
                setattr(self, k, self.json[k])

        self.df_factory = DfFactory(good_data_keys=self.good_data_keys())

    def __getitem__(self, key):
        "Trick to return any dict key as a property"
        # print(key, "==========")
        if key == "stats":
            key = "_stats" # syntactic sugar
        if key in self.json:
            return self.json[key]
        else:
            return [] # empty list so to support iteration tests / previously: False

    def __len__(self):
        "Return length of first object in JSON (skipping '_stats'"
        k = self.good_data_keys()
        try:
            return len(self.json[k[0]])
        except:
            return 0

    def good_data_keys(self,):
        "return the results keys other than stats"
        skips = ["_warnings", "_notes", "_stats", "_version"]
        return [x for x in self.json.keys() if x not in skips]

    def keys_and_count(self,):
        "Utility to preview contents of results object"
        return [(x, len(self.json[x])) for x in self.json.keys()]

    @property
    def count_total(self,):
        "Quickly return tot count for query (not for current payload)"
        if self.json.get("_stats"):
            return self.json['_stats']['total_count']
        else:
            return None

    @property
    def count_batch(self,):
        "Quickly return tot count for current batch from query"
        return len(self)

    @property
    def errors_string(self,):  # can't be called 'error' due to conflict with auto-set field
        "Quickly return errors string"
        if self.json.get("errors"):
            return self.json['errors']['query']['header'] + self.json['errors']['query']['details'][0]
        else:
            return ""

    def chunks(self, size=400, key=""):
        """
        Return an iterator for going through chunks of the JSON results. 
        NB the first available dict key is taken, to determine what is the data 
        object being returned. 
        Default size of chunks = 400 elements

        EG

        res = dslquery("search publications return publications limit 1000")
        test = [len(x) for x in res.chunks()]

        """

        if not key:
            if len(self.good_data_keys()) > 1:
                print(f"Please specify a key from {self.good_data_keys()}")
                return
            else:
                key = self.good_data_keys()[0]
        elif key not in self.good_data_keys():
            print(f"Invalid key: should be one of {self.good_data_keys()}")
            return 

        it = iter(self.json[key])
        chunk = list(islice(it, size))
        while chunk:
            yield chunk
            chunk = list(islice(it, size))

    # Dataframe Methods 

    def as_dataframe(self, key=""):
        "utility method: return inner json as a pandas dataframe"

        if not self.json.get("errors"):
            return self.df_factory.df_simple(self.json, key)


    def as_dataframe_authors(self):
        """Utility method: return inner json as a pandas dataframe, exposing authors + pubId
            Note: affiliations are not broken down. So one gets one row per author
        """
        if not self.json.get("errors"):
            return self.df_factory.df_authors(self.json)


    def as_dataframe_authors_affiliations(self):
        """Utility method: return inner json as a pandas dataframe, exposing authors + affiliations + pubId
        Affiliations ARE broken down and are returned as JSON - So one gets one row per affiliation (+1 row per author if having more than one affiliation)
        """
        if not self.json.get("errors"):
            return self.df_factory.df_authors_affiliations(self.json)

    def as_dataframe_concepts(self, key=""):
        """Utility method: return inner concepts list as a pandas dataframe, including several metrics for working with concepts.

        :param key: the JSON data key including concept information. If key is empty, the first available JSON key (eg 'publications') is used to build the dataframe

        The resulting dataframe has extra columns for ranking concepts:
        1) `concepts_count`: an integer representing the total number of concepts per document.
        2) `rank`: an integer representing the ranking of the concept within the list of concepts for a single document. E.g., the first concept has rank=1, while the fifth has rank=5.
        3) `score`: a float representing the importance of the concept in within a document. This is obtained by  normalizing its ranking against the total number of concepts for a single document (eg publication or grant). So if a document has 10 concepts in total, the first concept gets a score=1, the second score=0.9, etc..
        4) `frequency`: an integer representing how often that concept occurs within the full results-set returned by a query, i.e. how many documents have that concept name. So if a concept appears in 5 documents, frequency=5.
        5) `rank_avg`: the average (mean) value of all ranks for a single concept, across the full set of documents returned by the query. 
        6) `score_avg`: the average (mean) value of all scores for a single concept, across the full set of documents returned by a query.  
        7) `score_sum`: the sum of all scores for a single concept, across the full set of documents returned by a query.  
        """
        if not self.json.get("errors"):
            return self.df_factory.df_concepts(self.json, key)


    def as_dataframe_funders(self):
        """Utility method: return inner json as a pandas dataframe, for grants funders
        """
        if not self.json.get("errors"):
            return self.df_factory.df_grant_funders(self.json)

    def as_dataframe_investigators(self):
        """Utility method: return inner json as a pandas dataframe, for grants funders
        """
        if not self.json.get("errors"):
            return self.df_factory.df_grant_investigators(self.json)

    def as_dimensions_url(self, max=500, verbose=True):
        """Generates a valid Dimensions search URL using all the object IDs as filters.
        NOTE: this is EXPERIMENTAL and may break or be removed in future versions. 
        Also, doesn't work with all sources. 

        <max> 500 cause to prevent error 414 Request-URI Too Large 
        <verbose> to print out the warning

        General query structure for IDs: 
           
        `id: (pub.1120715293 OR pub.1120975084 OR pub.1122068834 OR pub.1120602308)`

        Final URL looks like this https://app.dimensions.ai/discover/publication?search_text=id%3A+%28pub.1120715293+OR+pub.1120975084+OR+pub.1122068834+OR+pub.1120602308%29

        """
        
        if verbose: print("Warning: this is an experimental and unsupported feature.")

        # hardcoded
        supported_url_templates = {
            'publications' : "https://app.dimensions.ai/discover/publication?search_text=",
            'grants' : "https://app.dimensions.ai/discover/grant?search_text=",
            'patents' : "https://app.dimensions.ai/discover/patent?search_text=",
            'clinical_trials' : "https://app.dimensions.ai/discover/clinical_trial?search_text=",
            'policy_documents' : "https://app.dimensions.ai/discover/policy_document?search_text=",
        }

        # just return first valid source found in results
        ids = []
        for sourcetype in supported_url_templates:
            if sourcetype in self.good_data_keys():
                try:
                    ids = [x['id'] for x in self.json[sourcetype]]
                    q = " OR ".join(ids)
                    if sourcetype == "grants":
                        q =  "grant_id: (" + q + ")"
                    else:
                        q =  "id: (" + q + ")"
                    q =  urllib.parse.quote_plus(q)
                    return supported_url_templates[sourcetype] + q
                except:
                    raise Exception("DslDataset records do not contain a valid ID field.")
        return None


    def __repr__(self):
        if self.json.get("errors"):
            return "<dimcli.DslDataset object #%s. Errors: %d>" % (str(id(self)), len(self.json['errors']))
        else:
            try:
                return "<dimcli.DslDataset object #%s. Records: %d/%d>" % (str(id(self)), self.count_batch, self.count_total)
            except:
                # non-search queries
                return "<dimcli.DslDataset object #%s. Dict keys: %s>" % (str(id(self)), ", ".join([f"'{x}'" for x in self.json]))





# 2019-12-17: for backward compatibility
# remove once all notebooks code has been updated 
Result = DslDataset
Dataset = DslDataset


#!/usr/bin/python
# -*- coding: utf-8 -*-

import requests
import json
from packaging import version
# try:
    
#     from packaging.version import parse
# except ImportError:
#     from pip._vendor.packaging.version import parse

from ..VERSION import __version__ as VERSION  # strip out the 'v'

URL_PATTERN = 'https://pypi.python.org/pypi/{package}/json'


def get_pypi_version(package, url_pattern=URL_PATTERN):
    """Return version of package on pypi.python.org using json."""
    req = requests.get(url_pattern.format(package=package))
    the_version = version.parse('0')
    if req.status_code == requests.codes.ok:
        j = json.loads(req.text.encode(req.encoding or "utf-8"))
        releases = j.get('releases', [])
        for release in releases:
            ver = version.parse(release)
            if not ver.is_prerelease:
                the_version = max(the_version, ver)
    return the_version



def is_dimcli_outdated(this_version=VERSION):
    "test if installed version of Dimcli is the latest one available on pypi"
    try:
        latest = get_pypi_version("dimcli")
        if version.parse(latest.base_version) > version.parse(this_version):
            return True
        else:
            return False
    except:
        return None



def print_dimcli_report(this_version=VERSION):
    "Util for CLI"
    try:
        latest = get_pypi_version("dimcli")
        print("====\nThe latest Dimcli version is ", latest.base_version )
        print("You have installed: ", VERSION)
        if version.parse(latest.base_version) > version.parse(this_version):
            print("====\nPlease upgrade: `pip install dimcli -U`")
        else:
            print("====\nYou're up to date.")
    except:
        print("Couldn't connect to the pypi server. Are you online?")


def print_dimcli_report_if_outdated(this_version=VERSION, force=False):
    "Util for checking latest Dimcli version - PS this version prints out a message only if Dimcli is out of date."
    try:
        latest = get_pypi_version("dimcli")
        if force or version.parse(latest.base_version) > version.parse(this_version):
            print("====\nHeads up! The latest Dimcli version is ", latest.base_version )
            print("You have installed: ", VERSION)
            print("====\nPlease upgrade: `pip install dimcli -U`")
    except:
        print("Couldn't connect to the pypi server. Are you online?")


#
# https://docs.dimensions.ai/dsl/data.html
#

# need a structure that contains desc / and possibly other lang metadata
# then maybe we can have 'children' as a key for nested objects

from .dsl_grammar_dict import *
from .dsl_grammar_categories import *

if True:
    vocab_data = GRAMMAR_DICT 
    syntax_data = SYNTAX_DICT 
    categories_data = CATEGORIES_DICT 
else:
    # @TODO get in real time from DSL
    pass 


# simple util that must stay here to avoid circular imports with utils.py
def merge_two_dicts(x, y):
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


class DslGrammar():
    """
    Wrapper for the DSL Grammar dict

    """
    def __init__(self, grammar_dict, extra_syntax, categories_dict=None):
        self.grammar = grammar_dict
        self.syntax = extra_syntax
        self.categories_names = list(categories_dict.keys())
        self.categories_data = categories_dict
        # DslGrammar.__init__(self, data)

    def __getitem__(self, key):
        "return dict key as slice"
        if key in self.grammar:
            return self.grammar[key]
        else:
            return False

    def __repr__(self):
        stats = [(x, len(self.grammar[x])) for x in self.grammar.keys()]
        return "<Dsl Grammar object #%s. %s>" % (str(id(self)), str(stats))

    # 
    # SYNTAX METHODS
    # 

    def allowed_starts(self, word=""):
        "Get a the allowed starts dict for Dimcli"
        allowed_starts = merge_two_dicts(self.allowed_starts_dsl_query(), self.allowed_starts_special_commands())
        if not word:
            return allowed_starts
        else:
            try:
                return allowed_starts[word]
            except:
                return []
    def allowed_starts_dsl_query(self):
        "Get a the allowed starts dict specific to he DSL syntax"
        return self.syntax['allowed_starts_dsl_query']
    def allowed_starts_special_commands(self):
        "Get a the allowed starts dict specific to he DIMCLI special syntax"
        return self.syntax['allowed_starts_special_commands']

    def lang(self):
        "Get a list of all lang operators"
        return self.syntax['lang_all']
    def lang_after_search(self):
        "Get a list of all lang operators after a valid search -source- statement"
        return self.syntax['lang_after_search']
    def lang_filter_operators(self):
        "Get a list of all lang operators for a filter statement eg  `search -source- where -filter-` statement"
        return self.syntax['lang_filter_operators']
    def lang_text_operators(self):
        "Get a list of all lang operators for a text search eg  `search -source- for 'x AND y'`"
        return self.syntax['lang_text_operators']
    def lang_after_for_text(self):
        "Get a list of all lang operators after a valid `for 'text'` statement"
        return self.syntax['lang_after_for_text']
    def lang_after_filter(self):
        "Get a list of all lang operators after a valid `search -source- where -filter=?-` statement"
        return self.syntax['lang_after_filter']
    def lang_after_sort_by(self):
        "Get a list of all lang operators after a valid `return -source- sort by -field-` statement"
        return self.syntax['lang_after_sort_by']
    def lang_after_return(self):
        "Get a list of all lang operators after a valid `return -source-` statement"
        return self.syntax['lang_after_return']
    def lang_after_limit(self):
        "Get a list of all lang operators after a valid `limit -N-` statement - ie skip"
        return self.syntax['lang_after_limit']

    def url_for_source(self, source):
        "Get a the Dimensions URL for a specific source"
        try:
            return self.syntax['dimensions_urls'][source]
        except:
            return []

    def object_id_patterns(self):
        """Return a dictionary representing known string-prefixes for object IDs
        Note: in some cases eg patents this is missing.
        """
        return self.syntax['dimensions_object_id_patterns']           
    # 
    # GRAMMAR METHODS IE the dsl sources / fields
    # 

    # generic

    def get_field_json(self, field, source=None, entity=None):
        "Get the raw json for a field"
        if source:
            try:
                return self.grammar['sources'][source]['fields'][field]
            except:
                return None
        elif entity:
            try:
                return self.grammar['entities'][entity]['fields'][field]
            except:
                return None            

    # sources 

    def sources(self):
        "Get a list of all sources available"
        return [x for x in self.grammar['sources'].keys()]

    def fields_for_source(self, source, filters=False, facets=False, fieldtype=False):
        "Get a list of all fields available"
        out= []
        if source not in self.sources():
            return []
        for field,specs in self.grammar['sources'][source]['fields'].items():
            if filters:
                if specs['is_filter']:
                    out.append(field)
            elif facets:
                if specs['is_facet']:
                    out.append(field)
            elif fieldtype:
                if specs['type'] == fieldtype:
                    out.append(field)
            else:
                out.append(field)
        return out
    
    def filters_for_source(self, source):
        "Get a list of all fields-filters available"
        return self.fields_for_source(source, filters=True)

    def facets_for_source(self, source):
        "Get a list of all fields-facets available"
        return self.fields_for_source(source, facets=True)

    def entity_type_for_source_facet(self, source, facet):
        "from a facet, return the entity name EG object link to entities"
        if not facet in self.facets_for_source(source):
            return []
        return self.grammar['sources'][source]['fields'][facet]['type']

    def desc_for_source_field(self, source, field):
        "from a source-field combination, return the description"
        if not field in self.fields_for_source(source):
            return ""
        return self.grammar['sources'][source]['fields'][field]['description']
    def desc_for_source_field_enriched(self, source, field):
        "A decription prefixed by extra info for a field eg if it is a facet"
        # desc = self.desc_for_source_field(source, field)
        json = self.get_field_json(field=field, source=source)
        try:
            desc = json['description'] or  ""
            if json['is_entity']:
                return "[ENTITY:%s] " % json['type'] + desc
            else:
                return "[LITERAL:%s] " % json['type'] + desc
        except:
            pass

    def fieldsets_for_source(self, source):
        "Get a list of all fieldsets available"
        try:
            return self.grammar['sources'][source]['fieldsets']
        except:
            return []

    def metrics_for_source(self, source):
        "Get a list of all metrics available"
        try:
            return [x for x in self.grammar['sources'][source]['metrics']]
        except:
            return []

    def search_fields_for_source(self, source):
        "Get a list of all search fields available"
        try:
            return self.grammar['sources'][source]['search_fields']
        except:
            return []

    # entity

    def entities(self):
        "Get a list of all entities available"
        return [x for x in self.grammar['entities'].keys()]

    def fields_for_entity(self, entity, filters=False, fieldtype=None):
        "Get a list of all fields available for an entity"
        out= []
        if entity not in self.entities():
            return []
        for field,specs in self.grammar['entities'][entity]['fields'].items():
            if filters:
                if specs['is_filter']:
                    out.append(field)
            elif fieldtype:
                if specs['type'] == fieldtype:
                    out.append(field)
            else:
                out.append(field)
        return out

    def filters_for_entity(self, source):
        "Get a list of all fields-filters available for an entity"
        return self.fields_for_entity(source, filters=True)

    def fields_for_entity_from_source_facet(self, source, facet):
        "From entity field-name used in a source, get the list of fields for that entity has."
        if facet not in self.facets_for_source(source):
            return []
        entity = self.entity_type_for_source_facet(source, facet)
        if entity == "researchers":
            # SPECIAL CASE / hardwired
            # researchers is used as an entity, but it's actually a source
            return self.filters_for_source('researchers')
        else:
            return self.filters_for_entity(entity)  # only filters as used in 'where' context
        
    def desc_for_entity_field(self, entity, field):
        "from a entity-field combination, return the description"
        if not field in self.fields_for_entity(entity):
            return ""
        return self.grammar['entities'][entity]['fields'][field]['description']
    def desc_for_entity_field_enriched(self, entity, field):
        "A decription prefixed by extra info for a field eg if it a string or number"
        # @TODO similar to desc_for_source_field_enriched // consider refactoring
        json = self.get_field_json(field=field, entity=entity)
        try:
            desc = json['description'] or  ""
            return "[LITERAL:%s] " % json['type'] + desc
        except:
            pass

    # research categories

    def categories(self, name=None):
        "Get a list of all research categories available"
        if not name:
            return self.categories_names
        else:
            if self.categories_data.get(name, ""):
                return [x['name'] for x in self.categories_data[name]]            


# init grammar object from data
G = DslGrammar(vocab_data, syntax_data, categories_data)





import configparser
import requests
import os.path
import os
import sys
import time
import json
import click

from .walkup import *


USER_DIR = os.path.expanduser("~/.dimensions/")
USER_CONFIG_FILE_NAME = "dsl.ini"
USER_CONFIG_FILE_PATH = os.path.expanduser(USER_DIR + USER_CONFIG_FILE_NAME)
USER_HISTORY_FILE = os.path.expanduser(USER_DIR + "history.txt")
#
USER_EXPORTS_DIR = os.path.expanduser("~/dimcli-exports/")
# _USER_JSON_OUTPUTS_DIR = os.path.expanduser(USER_EXPORTS_DIR + "json/")




CONNECTION = {'instance': None, 'url': None, 'username': None, 'password': None, 'key': None,  'token' : None}




def do_global_login(instance="live", username="", password="", key="", url="https://app.dimensions.ai"):
    "Login into DSL and set the connection object with token"
    
    global CONNECTION

    if not (username and password) and not key:
        # then use 'instance' shortcut to get local credentials
        fpath = get_init_file()
        config_section = read_init_file(fpath, instance)
        url = config_section['url']
        try:
            username = config_section['login']
            password = config_section['password']
        except:
            username, password = "", ""
        try:
            key = config_section['key']
        except:
            key = ""


    login_data = {'username': username, 'password': password, 'key': key}
    response = requests.post(
        '{}/api/auth.json'.format(url), json=login_data)
    response.raise_for_status()

    token = response.json()['token']

    CONNECTION['instance'] = instance
    CONNECTION['url'] = url
    CONNECTION['username'] = username
    CONNECTION['password'] = password
    CONNECTION['key'] = key
    CONNECTION['token'] = token



def refresh_login():
    """
    Method used to login again if the TOKEN has expired - using previously entered credentials
    """
    do_global_login(CONNECTION['instance'], CONNECTION['username'], CONNECTION['password'], CONNECTION['key'], CONNECTION['url'])


def reset_login():
    ""
    global CONNECTION
    CONNECTION = {'instance': None, 'url': None, 'username': None, 'password': None,  'key': None,  'token' : None}


def get_connection():
    global CONNECTION
    return CONNECTION


def is_logged_in():
    if CONNECTION['token']:
        return True
    else:
        print("Warning: you are not logged in. Please use `dimcli.login(username, password)` before querying.")
        return False


def get_init_file():
    """
    LOGIC

    a) if dsl.ini credentials file in WD that overrides everything
    b) try user level credentials ("~/.dimensions/") 
    c) try navigating up directory tree for dsl.ini

    => a) and c) are useful for jup notebooks without system wide installation
    => b) is the usual case for CLI

    ===================
    BACKGROUND 

    Authentication details can be stored in a `dsl.ini` file
    File contents need to have this structure:

    [instance.live]
    url=https://app.dimensions.ai
    login=your_username
    password=your_password
    key=your_key

    The section name has to start with "instance."
    Keyword "live" is the default name for most installations.

    If you have access to other Dimensions APIs just add an entry for them with a suitable name.
    ===================
    """
    if os.path.exists(os.getcwd() + "/" + USER_CONFIG_FILE_NAME):
        return os.getcwd() + "/" + USER_CONFIG_FILE_NAME
    elif os.path.exists(os.path.expanduser(USER_CONFIG_FILE_PATH )):
        return os.path.expanduser(USER_CONFIG_FILE_PATH )
    else:
        for c,d,f in walk_up(os.getcwd()):
            if os.path.exists(c + "/" + USER_CONFIG_FILE_NAME):
                return c + "/" + USER_CONFIG_FILE_NAME
    return None

def read_init_file(fpath, instance_name):
    """
    parse the credentials file
    """
    config = configparser.ConfigParser()
    try:
        config.read(fpath)
    except:
        click.secho(f"ERROR: `{USER_CONFIG_FILE_NAME}` credentials file not found." , fg="red")
        click.secho("HowTo: https://github.com/digital-science/dimcli#credentials-file", fg="red")
        sys.exit(0)
    # we have a good config file
    try:
        section = config['instance.' + instance_name]
    except:
        click.secho(f"ERROR: Credentials file `{fpath}` does not contain settings for instance: '{instance_name}''", fg="red")
        click.secho(f"Available instances are:")
        for x in config.sections():
            click.secho("'%s'" % x)
        click.secho("---\nSee Also: https://github.com/digital-science/dimcli#credentials-file")
        config.sections()
        sys.exit(0)
    return section





#!/usr/bin/python
# -*- coding: utf-8 -*-


"""
simple test queries [for DEVELOPMENT  / not part of official tests]

$ pip instal -e .
$ dimcli_quicktest 1

"""

import click 
import os
import requests

from .. import *
from ..shortcuts import *
from ..core.converters import *


@click.command()
@click.argument('test_number', nargs=1)
def main(test_number=1):
    
    login()
    dsl = Dsl()
    test_number = int(test_number)

    if test_number == 1:

        covid_q = '"2019-nCoV" OR "COVID-19" OR "SARS-CoV-2" OR (("coronavirus"  OR "corona virus") AND (Wuhan OR China))'
        q = f"""search publications 
                in full_data for "{dsl_escape(covid_q)}" 
                where year=2020 
            return publications[id+doi+pmid+pmcid+title+journal+publisher+mesh_terms+date+year+volume+issue+pages+open_access_categories+type+authors+research_orgs+funders+supporting_grant_ids+times_cited+altmetric+linkout] limit 1000"""

        q1 = """search publications for "malaria" return publications[id+authors]"""

        click.secho(q, fg="green")

        df = dslquery(q).as_dataframe()

        r = DfConverter(df, "publications")
        res = r.simplify_nested_objects()

        order = [   'Publication ID', 
                    'DOI',
                    'PMID',
                    'PMCID',
                    'Title',
                    'Abstract',
                    'Source title',
                    'Source UID',
                    'Publisher',
                    'MeSH terms',
                    'Publication Date',
                    'PubYear',
                    'Volume',
                    'Issue',
                    'Pagination',
                    'Open Access',
                    'Publication Type',
                    'Authors',
                    'Corresponding Author',
                    'Authors Affiliations',
                    'Research Organizations - standardized',
                    'GRID IDs',
                    'City of Research organization',
                    'Country of Research organization',
                    'Funder',
                    'UIDs of supporting grants',
                    'Times cited',
                    'Altmetric',
                    'Source Linkout',
                    'Dimensions URL']

        res = res[order]

        res.to_csv("test.csv", index=False)

    if test_number == 2:

        # click.secho("Testing as_dataframe_concepts on Publications data: ", fg="magenta")
        # res= dslquery("""search publications for "graphene" where year=2019 return publications[id+concepts+year+title+doi+journal] limit 1000""")
        # concepts = res.as_dataframe_concepts()
        # print(" ==> res.as_dataframe_concepts(): ", concepts)
        # concepts.info()
        # ----
        click.secho("Testing as_dataframe_concepts on Publications data - using SCORES: ", fg="magenta")
        res= dslquery("""search publications for "graphene" where year=2019 return publications[id+concepts_scores+year+title+doi+journal] limit 1000""")
        concepts = res.as_dataframe_concepts()
        print(" ==> res.as_dataframe_concepts(): ", concepts)
        concepts.info()
        # ----
        click.secho("Testing as_dataframe_concepts on Grants data: ", fg="magenta")
        res= dslquery("""search grants for "graphene" where active_year = 2019 return grants[basics+concepts] limit 500""")
        concepts = res.as_dataframe_concepts()
        print(" ==> res.as_dataframe_concepts(): ", concepts)
        concepts.info()
        click.secho("Completed test succesfully", fg="green")

if __name__ == '__main__':
    main()






#!/usr/bin/python
# -*- coding: utf-8 -*-

from prompt_toolkit.key_binding import KeyBindings

#
# KEY BINDINGS OVERRIDE
#

bindings = KeyBindings()


@bindings.add("c-space")
def _(event):
    """
    Start auto completion. If the menu is showing already, select the next
    completion.
    """
    b = event.app.current_buffer
    if b.complete_state:
        b.complete_next()
    else:
        b.start_completion(select_first=False)


@bindings.add("c-o")
def _(event):
    """
    Look up in docs
    """
    line = event.app.current_buffer.text
    if line:
        line_last_word = line.split()[-1]
        import webbrowser
        webbrowser.open("https://docs.dimensions.ai/dsl/search.html?q=" +
                        line_last_word)
    return



import json
import pandas as pd
try:
    from pandas import json_normalize
except:
    from pandas.io.json import json_normalize
from .utils import normalize_key, exists_key_in_dicts_list


class DfFactory(object):
    """
    Helper class containing methods for creating dataframes from API JSON

    The 'data' passed to its methods is always JSON/DICT data 
    """

    def __init__(self, good_data_keys=[]):
        self.good_keys = good_data_keys


    def df_simple(self, data, key):
        """Return inner json as a pandas dataframe
        If key is empty, the first available JSON key (eg 'publications') is used to determine
        what data should be turned into a dataframe. 
        """

        output = pd.DataFrame()
        valid_key = False

        if key and (key in self.good_keys):
            valid_key = key
        elif key and (key not in self.good_keys):
            print(f"[Warning] Dataframe cannot be created: invalid key. Should be one of {self.good_keys}")
        elif not key and self.good_keys:
            if len(self.good_keys) > 1:
                print(f"[Warning] Dataframe created from first available key, but more than one JSON key found: {self.good_keys}")
            valid_key = self.good_keys[0]
        else:
            pass 

        if valid_key:
            if data[valid_key] and type(data[valid_key]) == list:
                if type(data[valid_key][0]) == dict:
                    output = json_normalize(data[valid_key])
                else: # simple list of strings or numbers  
                    output = pd.DataFrame.from_dict(data[valid_key]) 
            else: # no list, then make one and try to return everything
                output = pd.DataFrame.from_dict([data])

        return output


    def df_authors(self, data):
        """Utility 
        Returns inner json as a pandas dataframe, exposing authors + pubId.
        List of affiliations per each author are not broken down and are returned as JSON. 
        So in essence you get one row per author.

        NOTE this method works only for publications searches -and it's clever enough to know if `authors` or `author_affiliations` (deprecated) field is used.

        Each publication.author_affiliations object has a nested list structure like this:
        ```
        [[{'first_name': 'Laura',
            'last_name': 'Pasin',
            'orcid': '',
            'current_organization_id': '',
            'researcher_id': '',
            'affiliations': [{'name': 'Department of Anesthesia and Intensive Care, Ospedale S. Antonio, Via Facciolati, 71, Padova, Italy'}]},
            {'first_name': 'Sabrina',
            'last_name': 'Boraso',
            'orcid': '',
            'current_organization_id': '',
            'researcher_id': '',
            'affiliations': [{'name': 'Department of Anesthesia and Intensive Care, Ospedale S. Antonio, Via Facciolati, 71, Padova, Italy'}]},
            {'first_name': 'Ivo',
            'last_name': 'Tiberio',
            'orcid': '',
            'current_organization_id': '',
            'researcher_id': '',
            'affiliations': [{'name': 'Department of Anesthesia and Intensive Care, Ospedale S. Antonio, Via Facciolati, 71, Padova, Italy'}]}]]
        ```

        """
        output = pd.DataFrame()

        if 'publications' in self.good_keys:

            if exists_key_in_dicts_list(data['publications'], "author_affiliations"):
                FIELD = "author_affiliations"
            elif exists_key_in_dicts_list(data['publications'], "authors"):
                FIELD = "authors"
            else:
                FIELD = ""

            if FIELD == "author_affiliations":
                # simplify deep nested dict structure for deprecated field 
                for x in data['publications']:
                    if 'author_affiliations' in x and x['author_affiliations']:  # if key exists and contents are not empty eg '[]'
                        if type(x['author_affiliations'][0]) == list: # then break down nested dict structure
                            x['author_affiliations'] = x['author_affiliations'][0]
                        elif type(x['author_affiliations'][0]) == dict: # = it's already been broken down
                            pass
                    else: # put in default empty element
                        x['author_affiliations'] = []
            elif FIELD == "authors":
                normalize_key("authors", data['publications'], [])
            
            if FIELD:
                output = json_normalize(data['publications'], record_path=[FIELD], meta=['id'], errors='ignore')
                output.rename(columns={"id": "pub_id"}, inplace=True)
        else:
            print(f"[Warning] Dataframe cannot be created as 'publications' were not found in data. Available: {self.good_keys}")
        return output


    def df_authors_affiliations(self, data):
        """Utility method
        return inner json as a pandas dataframe, exposing authors + affiliations + pubId
        affiliations ARE broken down and are returned as JSON 
        So one gets one row per affiliation (+1 row per author if having more than one affiliation)

        NOTE this method builds on self.df_authors()
        """
        
        authors = self.df_authors(data)
        if len(authors):
            affiliations = json_normalize(json.loads(authors.to_json(orient='records')), record_path=['affiliations'], 
               meta=['pub_id', 'researcher_id', 'first_name', 'last_name'], record_prefix='aff_')
        else:
            affiliations = authors # empty df
        affiliations.fillna('', inplace=True) # 2019-09-30: simplifies subsequent operations
        return affiliations




    def df_concepts(self, data, key):
        """from a list of publications or grants including concepts, return a DF with one line per concept
        Enrich the dataframe with scores and other metrics.
        """

        FIELD_NAME = "concepts"
        FIELD_NAME_SCORES = "concepts_scores"
        ROUNDING = 5

        if not ('publications' in self.good_keys) and not ('grants' in self.good_keys): 
            s = f"Dataframe can be created only with searches returning 'publications' or 'grants' . Available: {self.good_keys}"
            raise Exception(s)

        concepts = self.df_simple(data, key)

        if (FIELD_NAME not in concepts.columns) and (FIELD_NAME_SCORES not in concepts.columns):
            s = f"Dataframe requires raw concepts data, but no 'concepts' or 'concepts_scores' column was not found in: {concepts.columns.to_list()}"
            raise Exception(s)             

        if not 'id' in concepts.columns:
            s = f"Dataframe requires an 'id' column for counting concepts, which was not found in: {concepts.columns.to_list()}"
            raise Exception(s)   

        if FIELD_NAME_SCORES in concepts.columns:
            # use `concepts_scores` field preferably

            df = concepts.explode(FIELD_NAME_SCORES)
            df.dropna(subset=[FIELD_NAME_SCORES], inplace=True)  # remove rows if there is no concept
            df.reset_index(inplace=True, drop=True)
            original_cols = [x for x in df.columns.to_list() if x != FIELD_NAME_SCORES]
            df = df.drop(FIELD_NAME_SCORES, 1).assign(**pd.json_normalize(df[FIELD_NAME_SCORES]))  # unpack dict with new columns
            df = df[df.relevance != 0]  # remove 0-relevance scores
            df['relevance'] = df['relevance'].round(ROUNDING)
            df.rename(columns={"relevance": "score"}, inplace=True) 
            df['frequency'] = df.groupby('concept')['concept'].transform('count')
            df['concepts_count'] = df.groupby("id")['concept'].transform('size')
  
        else: 
            # with traditional 'concepts', scores are simulated

            df = concepts.explode(FIELD_NAME)
            original_cols = [x for x in df.columns.to_list() if x != FIELD_NAME]
            df.dropna(subset=[FIELD_NAME], inplace=True) # remove rows if there is no concept
            df.rename(columns={FIELD_NAME: "concept"}, inplace=True)
            df['frequency'] = df.groupby('concept')['concept'].transform('count')
            df['concepts_count'] = df.groupby("id")['concept'].transform('size')
            ranks = df.groupby('id').cumcount()+1
            # scores = normalized rank from 0 to 1, where 1 is the highest rank
            df['score'] = ((df['concepts_count']+1) - ranks) / df['concepts_count']
            df['score'] = df['score'].round(ROUNDING)

        # finally
        df['score_avg'] = df.groupby('concept')['score'].transform('mean').round(ROUNDING)
        df.reset_index(drop=True, inplace=True)

        out_cols = original_cols + ['concepts_count', 'concept', 'score', 'frequency', 'score_avg' ]
        return df[out_cols]




    def _df_concepts(self, data, key):
        """from a list of publications or grants including concepts, return a DF with one line per concept
        Enrich the dataframe with scores and other metrics.
        """

        FIELD_NAME = "concepts"
        ROUNDING = 5

        if not ('publications' in self.good_keys) and not ('grants' in self.good_keys): 
            s = f"Dataframe can be created only with searches returning 'publications' or 'grants' . Available: {self.good_keys}"
            raise Exception(s)

        concepts = self.df_simple(data, key)

        if not 'concepts' in concepts.columns:
            s = f"Dataframe requires raw concepts data, but a 'concepts' column was not found in: {concepts.columns.to_list()}"
            raise Exception(s)             

        if not 'id' in concepts.columns:
            s = f"Dataframe requires an 'id' column for counting concepts, which was not found in: {concepts.columns.to_list()}"
            raise Exception(s)   

        df = concepts.explode(FIELD_NAME)
        original_cols = [x for x in df.columns.to_list() if x != FIELD_NAME]
        df.rename(columns={FIELD_NAME: "concept"}, inplace=True)
        df.dropna(inplace=True) # remove rows if there is no concept
        df['frequency'] = df.groupby('concept')['concept'].transform('count')
        df['concepts_count'] = df.groupby("id")['concept'].transform('size')
        ranks = df.groupby('id').cumcount()+1
        # scores = normalized rank from 0 to 1, where 1 is the highest rank
        df['score'] = ((df['concepts_count']+1) - ranks) / df['concepts_count']
        df['score'] = df['score'].round(ROUNDING)

        df['score_avg'] = df.groupby('concept')['score'].transform('mean').round(ROUNDING)

        df.reset_index(drop=True, inplace=True)

        out_cols = original_cols + ['concepts_count', 'concept', 'score', 'frequency', 'score_avg' ]
        return df[out_cols]


    def df_grant_funders(self, data):
        """Utility method: return inner json as a pandas dataframe, for grants funders
        """
        output = pd.DataFrame()
        FIELD = "funders"

        if 'grants' in self.good_keys:    
            normalize_key(FIELD, data['grants'], [])   
            output = json_normalize(data['grants'], record_path=[FIELD], meta=['id', 'title', 'start_date', 'end_date'], meta_prefix="grant_", errors='ignore')
        else:
            print(f"[Warning] Dataframe cannot be created as 'grants' were not found in data. Available: {self.good_keys}")

        return output

    def df_grant_investigators(self, data):
        """Utility method: return inner json as a pandas dataframe, for grants investigators 
        """
        output = pd.DataFrame()
        FIELD = "investigator_details"

        if 'grants' in self.good_keys:    
            normalize_key(FIELD, data['grants'], [])
            output = json_normalize(data['grants'], record_path=[FIELD], meta=['id', 'title', 'start_date', 'end_date'], meta_prefix="grant_", errors='ignore')
        else:
            print(f"[Warning] Dataframe cannot be created as 'grants' were not found in data. Available: {self.good_keys}")

        return output

    def df_grant_investigators_affiliations(self, data):
        """Utility method
        return inner json as a pandas dataframe, exposing investigators + affiliations + pubId
        affiliations ARE broken down and are returned as JSON 
        So one gets one row per affiliation (+1 row per investigators if having more than one affiliation)

        NOTE this method builds on self.df_grant_investigators()
        """
        
        investigators = self.df_grant_investigators(data)
        if len(investigators):
            affiliations = json_normalize(json.loads(investigators.to_json(orient='records')), record_path=['affiliations'], 
               meta=['id', 'first_name', 'last_name',  'role', 'grant_id', 'grant_title', 'grant_start_date', 'grant_end_date' ], record_prefix='aff_')
        else:
            affiliations = authors # empty df
        affiliations.fillna('', inplace=True) # 2019-09-30: simplifies subsequent operations
        return affiliations





#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import click
from pprint import pprint
import webbrowser

from .VERSION import *

from .core.auth import USER_DIR, USER_CONFIG_FILE_PATH, USER_HISTORY_FILE
from .core.api import *
from .core.utils import open_multi_platform, init_config_folder, print_warning_prompt_version, preview_contents, print_dimensions_url
from .core.version_utils import print_dimcli_report, is_dimcli_outdated

try:
    from prompt_toolkit.formatted_text import HTML
    from prompt_toolkit.shortcuts import CompleteStyle
    from prompt_toolkit import PromptSession
    from prompt_toolkit.styles import Style
    PROMPT_TOOLKIT_VERSION_OK = True
except:
    PROMPT_TOOLKIT_VERSION_OK = False # => repl disabled


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("instance_name", nargs=1, default="live")
@click.option(
    "--init",
    is_flag=True,
    help="Create a configuration file with your API credentials.")
@click.option(
    "--settings",
    is_flag=True,
    help="Show the local configuration file.")
@click.option(
    "--vcheck",
    is_flag=True,
    help="Check online if your dimcli version is the latest.")
@click.option(
    "--history", is_flag=True, help="Open history file with default editor.")
@click.option(
    "--id", "-id", help="Resolve a Dimensions ID to its public URL.")
@click.option(
    "--websearch", "-w", help="Search a quoted string in Dimensions web.")
@click.pass_context
def main_cli(ctx, instance_name=None, init=False, settings=False, vcheck=False, history=False, id=None, websearch=None):
    """
    Python client for the Dimensions Analytics API.
    More info: https://github.com/digital-science/dimcli
    """
    click.secho("Dimcli - Dimensions API Client (" + VERSION + ")", dim=True)

    if init:
        init_config_folder(USER_DIR, USER_CONFIG_FILE_PATH)
        return

    if vcheck:
        print_dimcli_report()
        return

    if id:
        print_dimensions_url(id)
        return 

    if websearch:
        url = dimensions_url_search(websearch)
        webbrowser.open(url)
        return 

    if not os.path.exists(USER_CONFIG_FILE_PATH):
        click.secho(
            "Credentials file not found - you can create one by typing: `dimcli --init`",
            fg="red",
        )
        click.secho(
            "More info: https://github.com/digital-science/dimcli#credentials-file",
            dim=True,
        )
        return

    if settings:
        preview_contents(USER_CONFIG_FILE_PATH)
        return

    if history:
        if os.path.exists(USER_HISTORY_FILE):
            open_multi_platform(USER_HISTORY_FILE)
        return

    if PROMPT_TOOLKIT_VERSION_OK:
        from .repl import repl
        # try online version check
        test = is_dimcli_outdated()
        if test:
            click.secho("====")
            click.secho("Heads up: there is a newer version of Dimcli available.", bold=True)
            click.secho("Update with `pip install dimcli -U` or, for more info, visit https://pypi.org/project/dimcli .\n====")
        # launch REPL
        repl.run(instance_name)
    else:
        print_warning_prompt_version()



if __name__ == "__main__":
    main_cli()



from .VERSION import __version__, VERSION

from .core.api import Dsl, DslDataset
from .core.dsl_grammar import G 
from .core.version_utils import print_dimcli_report_if_outdated

import click

try:
    # if run outside iPython, the magic fails so we use this as a test
    # https://stackoverflow.com/questions/32538758/nameerror-name-get-ipython-is-not-defined
    get_ipython()
    ipython_env = True
except:
    ipython_env = False

if ipython_env:
    from .jupyter import magics


#
# determine if we are in Google Colab or Jupyter
try:
    from google.colab import files
    # load table plugin => %load_ext google.colab.data_table
    get_ipython().run_line_magic("load_ext", "google.colab.data_table")
    COLAB_ENV = True
except:
    COLAB_ENV = False



def login(username="", password="", endpoint="https://app.dimensions.ai", instance="live", key="", verbose=True):
    """
    Login into the Dimensions API and obtain a query token. 
    - If credentials are not passed, we attempt to login using the local dsl.ini credentials file. 
    - If COLAB is detected and user/psw is not available, try running the interactive login

    Args: 
    * username
    * password
    * endpoint (defaults to "https://app.dimensions.ai")
    * instance (defaults to "live")
    * key (for newest login mechanism)
    """
    from .core.auth import do_global_login, get_connection

    try:
        do_global_login(instance, username, password, key, endpoint)
    except Exception as e:
        print("Login failed: please ensure your credentials are correct.")
        raise(e)

    CONNECTION = get_connection()

    if CONNECTION['token'] and verbose:
        _print_login_success(CONNECTION, username, password, key)
        print_dimcli_report_if_outdated()



def _print_login_success(CONNECTION, username, password, key):
    click.secho("Dimcli - Dimensions API Client (" + VERSION + ")", dim=True)
    CLIENT = Dsl(verbose=False)
    # dynamically retrieve dsl version 
    try:
        _info = CLIENT.query("describe version")['release']
    except:
        _info = "not available"

    if (username and password) or key:
        _method = "manual login"
    else:
        _method = "dsl.ini file"
    click.secho(f"Connected to: {CLIENT._url} - DSL v{_info}", dim=True)
    click.secho(f"Method: {_method}", dim=True)




def logout():
    """
    Reset the connectiont to the Dimensions API. This allows to create a new connection subsequently, eg to a different endpoint.
    """
    from .core.auth import reset_login, get_connection
    CONNECTION = get_connection()
    if CONNECTION['token']:
        reset_login()
        print("Logging out... done") 
    else:
        print("Please login first") 



def login_status():
    """Simply output info on whether we are logged in or not"""
    from .core.auth import get_connection
    CONNECTION = get_connection()
    if CONNECTION['token']:
        print("Dimcli %s - Succesfully connected to <%s>" % (str(VERSION), CONNECTION['url'])) 
        return True
    else:
        print("Status: not logged in") 
        return False